#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2016-2020, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

#pylint: disable-msg=empty-docstring
#pylint: disable-msg=missing-docstring
#pylint: disable-msg=invalid-name
#pylint: disable-msg=global-statement

import logging
from collections import deque
from os import environ
from sys import modules, stdout
from time import sleep
from unittest import TestCase, TestSuite, SkipTest, makeSuite, main as ut_main
from pyftdi import FtdiLogger
from pyftdi.ftdi import Ftdi
from pyftdi.gpio import (GpioAsyncController,
                         GpioSyncController,
                         GpioMpsseController)
from pyftdi.misc import to_bool

# When USB virtualization is enabled, this loader is instanciated
VirtLoader = None


class FtdiTestCase(TestCase):
    """Common features for all tests.
    """

    @classmethod
    def setUpClass(cls):
        cls.debug = to_bool(environ.get('FTDI_DEBUG', 'off'), permissive=False)
        cls.url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        cls.loader = None

    @classmethod
    def tearDownClass(cls):
        if cls.loader:
            cls.loader.unload()

    def setUp(self):
        if self.debug:
            print('.'.join(self.id().split('.')[-2:]))


class GpioAsyncTestCase(FtdiTestCase):
    """FTDI Asynchronous GPIO driver test case.

       Please ensure that the HW you connect to the FTDI port A does match
       the encoded configuration. Check your HW setup before running this test
       as it might damage your HW. You've been warned.

       Low nibble is used as input, high nibble is used as output. They should
       be interconnected as follow:

       * b0 should be connected to b4
       * b1 should be connected to b5
       * b2 should be connected to b6
       * b3 should be connected to b7

       Note that this test cannot work with LC231X board, as FTDI is stupid
       enough to add ubidirectionnal output buffer on DTR and RTS, so both
       nibbles have at lest on output pin...
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        if VirtLoader:
            cls.loader = VirtLoader()
            with open('pyftdi/tests/resources/ft232r.yaml', 'rb') as yfp:
                cls.loader.load(yfp)
            vftdi = cls.loader.get_virtual_ftdi(1, 1)
            vport = vftdi.get_port(1)
            # create virtual connections as real HW
            in_pins = [vport[pos] for pos in range(4)]
            out_pins = [vport[pos] for pos in range(4, 8)]
            for in_pin, out_pin in zip(in_pins, out_pins):
                out_pin.connect_to(in_pin)
        if cls.url == 'ftdi:///1':
            # assumes that if not specific device is used, and a multiport
            # device is connected, there is no loopback wires between pins of
            # the same port. This hack allows to run the same test with a
            # FT232H, then a FT2232H for ex, to that with two test sessions
            # the whole test set is run. If a specific device is selected
            # assume the HW always match the expected configuration.
            ftdi = Ftdi()
            ftdi.open_from_url(cls.url)
            count = ftdi.device_port_count
            ftdi.close()
            cls.skip_loopback = count > 1
        else:
            cls.skip_loopback = False

    def test_gpio_freeze(self):
        """Simple test to demonstrate freeze on close.

           For now, it requires a logic analyzer to verify the output,
           this is not automatically validated by SW
        """
        direction = 0xFF & ~((1 << 4) - 1) # 4 Out, 4 In
        gpio = GpioAsyncController()
        gpio.configure(self.url, direction=direction, frequency=1e3,
                       initial=0x0)
        port = gpio.get_gpio()
        # emit a sequence as a visual marker on b3,b2,b1,b0
        port.write([x<<4 for x in range(16)])
        sleep(0.01)
        # write 0b0110 to the port
        port.write(0x6<<4)
        sleep(0.001)
        # close w/o freeze: all the outputs should be reset (usually 0b1111)
        # it might need pull up (or pull down) to observe the change as
        # output are not high-Z.
        gpio.close()
        sleep(0.01)
        gpio.configure(self.url, direction=direction, frequency=1e3,
                       initial=0x0)
        port = gpio.get_gpio()
        # emit a sequence as a visual marker with on b3 and b1
        port.write([(x<<4)&0x90 for x in range(16)])
        sleep(0.01)
        # write 0b0110 to the port
        port.write(0x6<<4)
        sleep(0.01)
        # close w/ freeze: outputs should not be reset (usually 0b0110)
        gpio.close(True)


    def test_gpio_values(self):
        """Simple test to demonstrate bit-banging.
        """
        if self.skip_loopback:
            raise SkipTest('Skip loopback test on multiport device')
        direction = 0xFF & ~((1 << 4) - 1) # 4 Out, 4 In
        gpio = GpioAsyncController()
        gpio.configure(self.url, direction=direction, frequency=1e6,
                       initial=0x0)
        port = gpio.get_gpio()  # useless, for API duck typing
        # legacy API: peek mode, 1 byte
        ingress = port.read()
        self.assertIsInstance(ingress, int)
        # peek mode always gives a single byte output
        ingress = port.read(peek=True)
        self.assertIsInstance(ingress, int)
        # stream mode always gives a bytes buffer
        port.write([0xaa for _ in range(256)])
        ingress = port.read(100, peek=False, noflush=False)
        self.assertIsInstance(ingress, bytes)
        if not VirtLoader:
            # the virtual task may sometimes not be triggered soon enough
            self.assertGreater(len(ingress), 2)
        # direct mode is not available with multi-byte mode
        self.assertRaises(ValueError, port.read, 3, True)
        ingress = port.read(3)
        self.assertIsInstance(ingress, bytes)
        if not VirtLoader:
            # the virtual task may sometimes not be triggered soon enough
            self.assertGreater(len(ingress), 0)
        self.assertLessEqual(len(ingress), 3)
        port.write(0x00)
        port.write(0xFF)
        # only 8 bit values are accepted
        self.assertRaises(ValueError, port.write, 0x100)
        port.write([0x00, 0xFF, 0x00])
        port.write(bytes([0x00, 0xFF, 0x00]))
        # only 8 bit values are accepted
        self.assertRaises(ValueError, port.write, [0x00, 0x100, 0x00])
        # check direction API
        port.set_direction(0xFF, 0xFF & ~direction)
        gpio.close()

    def test_gpio_initial(self):
        """Check initial values.
        """
        if self.skip_loopback:
            raise SkipTest('Skip initial test on multiport device')
        if not self.loader:
            raise SkipTest('Skip initial test on physical device')
        direction = 0xFF & ~((1 << 4) - 1) # 4 Out, 4 In
        vftdi = self.loader.get_virtual_ftdi(1, 1)
        vport = vftdi.get_port(1)
        gpio = GpioAsyncController()
        for initial in (0xaf, 0xf0, 0x13, 0x00):
            gpio.configure(self.url, direction=direction, frequency=1e6,
                           initial=initial)
            expect = (initial & 0xF0) | (initial >> 4)
            self.assertEqual(vport.gpio, expect)
            gpio.close()

    def test_gpio_loopback(self):
        """Check I/O.
        """
        if self.skip_loopback:
            raise SkipTest('Skip loopback test on multiport device')
        gpio = GpioAsyncController()
        direction = 0xFF & ~((1 << 4) - 1) # 4 Out, 4 In
        gpio.configure(self.url, direction=direction, frequency=800000)
        for out in range(16):
            # print(f'Write {out:04b} -> {out << 4:08b}')
            gpio.write(out << 4)
            fback = gpio.read()
            lsbs = fback & ~direction
            msbs = fback >> 4
            # check inputs match outputs
            self.assertEqual(lsbs, out)
            # check level of outputs match the ones written
            self.assertEqual(msbs, out)
        outs = list([(out & 0xf) << 4 for out in range(1000)])
        gpio.write(outs)
        gpio.ftdi.read_data(512)
        for _ in range(len(outs)):
            _ = gpio.read(14)
        last = outs[-1] >> 4
        for _ in range(10):
            fbacks = gpio.read(1000)
            for fback in fbacks:
                lsbs = fback & ~direction
                msbs = fback >> 4
                # check inputs match last output
                self.assertEqual(lsbs, last)
                # check level of output match the last written
                self.assertEqual(msbs, last)
        gpio.close()

    def test_gpio_baudate(self):
        # this test requires an external device (logic analyser or scope) to
        # check the bitbang read and bitbang write signal (BB_RD, BB_WR) and
        # mesure their frequency. The EEPROM should be configured to enable
        # those signal on some of the CBUS pins, for example.
        gpio = GpioAsyncController()
        direction = 0xFF & ~((1 << 4) - 1) # 4 Out, 4 In
        gpio.configure(self.url, direction=direction)
        buf = bytes([0xf0, 0x00, 0xf0, 0x00, 0xf0, 0x00, 0xf0, 0x00])
        freqs = [50e3, 200e3, 1e6, 3e6]
        if gpio.ftdi.is_H_series:
            freqs.extend([6e6, 10e6, 12e6])
        gpio.read(128)
        for freq in freqs:
            # set the bitbang refresh rate
            gpio.set_frequency(freq)
            self.assertEqual(gpio.frequency, freq)
            # be sure to leave enough time to purge buffers (HW FIFO) or
            # the frequency changes occur on the current buffer...
            gpio.write(buf)
            gpio.read(128)
            sleep(0.01)
        gpio.close()


class GpioSyncTestCase(FtdiTestCase):
    """FTDI Synchrnous GPIO driver test case.

       Please ensure that the HW you connect to the FTDI port A does match
       the encoded configuration. Check your HW setup before running this test
       as it might damage your HW. You've been warned.

       Low nibble is used as input, high nibble is used as output. They should
       be interconnected as follow:

       * b0 should be connected to b4
       * b1 should be connected to b5
       * b2 should be connected to b6
       * b3 should be connected to b7

       Note that this test cannot work with LC231X board, as FTDI is stupid
       enough to add ubidirectionnal output buffer on DTR and RTS, so both
       nibbles have at lest on output pin...
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        if VirtLoader:
            cls.loader = VirtLoader()
            with open('pyftdi/tests/resources/ft232r.yaml', 'rb') as yfp:
                cls.loader.load(yfp)
            vftdi = cls.loader.get_virtual_ftdi(1, 1)
            vport = vftdi.get_port(1)
            # create virtual connections as real HW
            in_pins = [vport[pos] for pos in range(4)]
            out_pins = [vport[pos] for pos in range(4, 8)]
            for in_pin, out_pin in zip(in_pins, out_pins):
                out_pin.connect_to(in_pin)
        if cls.url == 'ftdi:///1':
            # assumes that if not specific device is used, and a multiport
            # device is connected, there is no loopback wires between pins of
            # the same port. This hack allows to run the same test with a
            # FT232H, then a FT2232H for ex, to that with two test sessions
            # the whole test set is run. If a specific device is selected
            # assume the HW always match the expected configuration.
            ftdi = Ftdi()
            ftdi.open_from_url(cls.url)
            count = ftdi.device_port_count
            ftdi.close()
            cls.skip_loopback = count > 1
        else:
            cls.skip_loopback = False

    def test_gpio_values(self):
        """Simple test to demonstrate bit-banging.
        """
        if self.skip_loopback:
            raise SkipTest('Skip loopback test on multiport device')
        direction = 0xFF & ~((1 << 4) - 1) # 4 Out, 4 In
        gpio = GpioSyncController()
        gpio.configure(self.url, direction=direction, initial=0xee)
        outs = bytes([(out & 0xf)<<4 for out in range(1000)])
        ins = gpio.exchange(outs)
        exp_in_count = min(len(outs), gpio.ftdi.fifo_sizes[0])
        self.assertEqual(len(ins), exp_in_count)
        last = None
        for sout, sin in zip(outs, ins):
            if last is not None:
                # output nibble
                sin_out = sin >> 4
                # input nibble
                sin_in = sin & 0xF
                # check inputs match last output
                self.assertEqual(sin_out, last)
                # check level of output match the last written
                self.assertEqual(sin_in, last)
            # an IN sample if captured on the next clock of the OUT sample
            # keep the MSB nibble, i.e. the nibble configured as output
            last = sout >> 4
        gpio.close()

    def test_gpio_baudate(self):
        # this test requires an external device (logic analyser or scope) to
        # check the bitbang read and bitbang write signal (BB_RD, BB_WR) and
        # mesure their frequency. The EEPROM should be configured to enable
        # those signal on some of the CBUS pins, for example.
        gpio = GpioSyncController()
        direction = 0xFF & ~((1 << 4) - 1) # 4 Out, 4 In
        gpio.configure(self.url, direction=direction)
        buf = bytes([0xf0, 0x00] * 64)
        freqs = [50e3, 200e3, 1e6, 3e6]
        if gpio.ftdi.is_H_series:
            freqs.extend([6e6, 10e6, 12e6])
        for freq in freqs:
            # set the bitbang refresh rate
            gpio.set_frequency(freq)
            self.assertEqual(gpio.frequency, freq)
            # be sure to leave enough time to purge buffers (HW FIFO) or
            # the frequency changes occur on the current buffer...
            gpio.exchange(buf)
            sleep(0.01)
        gpio.close()


class GpioMultiportTestCase(FtdiTestCase):
    """FTDI GPIO test for multi-port FTDI devices, i.e. FT2232H/FT4232H.

       Please ensure that the HW you connect to the FTDI port A does match
       the encoded configuration. Check your HW setup before running this test
       as it might damage your HW. You've been warned.

       First port is used as input, second port is used as output. They should
       be interconnected as follow:

       * AD0 should be connected to BD0
       * AD1 should be connected to BD1
       * AD2 should be connected to BD2
       * AD3 should be connected to BD3
       * AD0 should be connected to BD0
       * AD1 should be connected to BD1
       * AD2 should be connected to BD2
       * AD3 should be connected to BD3
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        if VirtLoader:
            cls.loader = VirtLoader()
            with open('pyftdi/tests/resources/ft2232h.yaml', 'rb') as yfp:
                cls.loader.load(yfp)
            vftdi = cls.loader.get_virtual_ftdi(1, 1)
            vport1 = vftdi.get_port(1)
            vport2 = vftdi.get_port(2)
            # create virtual connections as real HW
            in_pins = [vport1[pos] for pos in range(8)]
            out_pins = [vport2[pos] for pos in range(8)]
            for in_pin, out_pin in zip(in_pins, out_pins):
                out_pin.connect_to(in_pin)
        ftdi = Ftdi()
        ftdi.open_from_url(cls.url)
        count = ftdi.device_port_count
        pos = ftdi.port_index
        ftdi.close()
        if pos != 1:
            raise ValueError("FTDI interface should be the device's first")
        if count < 2:
            raise SkipTest('FTDI device is not a multi-port device')
        url = cls.url[:-1]
        cls.urls = [f'{url}1', f'{url}2']

    def test_gpio_peek(self):
        """Check I/O.
        """
        gpio_in, gpio_out = GpioAsyncController(), GpioAsyncController()
        gpio_in.configure(self.urls[0], direction=0x00, frequency=1e6)
        gpio_out.configure(self.urls[1], direction=0xFF, frequency=1e6)
        for out in range(256):
            gpio_out.write(out)
            outv = gpio_out.read()
            inv = gpio_in.read()
            # check inputs match outputs
            self.assertEqual(inv, out)
            # check level of outputs match the ones written
            self.assertEqual(outv, out)
        gpio_in.close()
        gpio_out.close()

    def test_gpio_stream(self):
        """Check I/O streaming
        """
        if VirtLoader:
            # this would require to synchronize virtual clock between all ports
            # which is not supported by the virtual framework
            raise SkipTest('Skip gpio stream with virtual device')
        gpio_in, gpio_out = GpioAsyncController(), GpioAsyncController()
        gpio_in.configure(self.urls[0], direction=0x00, frequency=1e4)
        gpio_out.configure(self.urls[1], direction=0xFF, frequency=1e4)
        outs = bytes(range(256))
        gpio_out.write(outs)
        # read @ same speed (and same clock source, so no jitter), flushing
        # the byffer which has been filled since the port has been opened
        ins = gpio_in.read(len(outs))
        qout = deque(outs)
        ifirst = ins[0]
        # the inout stream should be a copy of the output stream, minus a
        # couple of missing samples that did not get captured while output
        # was streaming but read command has not been yet received.
        while qout:
            if qout[0] == ifirst:
                break
            qout.popleft()
        # offset is the count of missed bytes
        offset = len(ins)-len(qout)
        self.assertGreater(offset, 0) # no more output than input
        self.assertLess(offset, 16) # seems to be in the 6..12 range
        # print('Offset', offset)
        # check that the remaining sequence match
        for sout, sin in zip(qout, ins):
            #print(f'{sout:08b} --> {sin:08b}')
            # check inputs match outputs
            self.assertEqual(sout, sin)
        gpio_in.close()
        gpio_out.close()

    def test_direction(self):
        b5 = 1 << 5
        for controller in (GpioAsyncController,
                           GpioSyncController,
                           GpioMpsseController):
            gpio_in, gpio_out = controller(), controller()
            gpio_in.configure(self.urls[0], direction=0x00, frequency=1e6,
                              debug=self.debug)
            gpio_out.configure(self.urls[1], direction=0xFF, frequency=1e6,
                               debug=self.debug)
            for direction in None, 0xFF, b5, 0xF0:
                if direction is not None:
                    gpio_out.set_direction(0xFF, direction)
                for out in 0 << 5, 1 << 5:
                    if controller != GpioSyncController:
                        gpio_out.write(out)
                        outp = gpio_out.read(1)
                        inp = gpio_in.read(1)
                        if controller == GpioMpsseController:
                            outp = outp[0]
                            inp = inp[0]
                    else:
                        # write twice the output value, only the second value
                        # matters (see Sync bitbang for details)
                        outp = gpio_out.exchange(bytes([out, out]))[1]
                        # write anything as we just need the input value which
                        # is sampled on each write
                        inp = gpio_in.exchange(b'\x00')[0]
                    self.assertEqual(outp & b5, out)
                    self.assertEqual(inp & b5, out)


class GpioMpsseTestCase(FtdiTestCase):
    """FTDI GPIO test for 16-bit port FTDI devices, i.e. FT2232H.

       Please ensure that the HW you connect to the FTDI port A does match
       the encoded configuration. Check your HW setup before running this test
       as it might damage your HW. You've been warned.

       First port is used as input, second port is used as output. They should
       be interconnected as follow:

       * AD0 should be connected to BD0
       * AD1 should be connected to BD1
       * AD2 should be connected to BD2
       * AD3 should be connected to BD3
       * AD0 should be connected to BD0
       * AD1 should be connected to BD1
       * AD2 should be connected to BD2
       * AD3 should be connected to BD3
       * AC0 should be connected to BC0
       * AC1 should be connected to BC1
       * AC2 should be connected to BC2
       * AC3 should be connected to BC3
       * AC0 should be connected to BC0
       * AC1 should be connected to BC1
       * AC2 should be connected to BC2
       * AC3 should be connected to BC3
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        if VirtLoader:
            cls.loader = VirtLoader()
            with open('pyftdi/tests/resources/ft2232h.yaml', 'rb') as yfp:
                cls.loader.load(yfp)
            vftdi = cls.loader.get_virtual_ftdi(1, 1)
            vport1 = vftdi.get_port(1)
            vport2 = vftdi.get_port(2)
            # create virtual connections as real HW
            in_pins = [vport1[pos] for pos in range(16)]
            out_pins = [vport2[pos] for pos in range(16)]
            for in_pin, out_pin in zip(in_pins, out_pins):
                out_pin.connect_to(in_pin)
            # prevent from using the tracer twice (Ftdi & VirtualFtdi)
            cls.debug_mpsse = False
        else:
            cls.debug_mpsse = cls.debug
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        ftdi = Ftdi()
        ftdi.open_from_url(url)
        count = ftdi.device_port_count
        width = ftdi.port_width
        ftdi.close()
        if count < 2:
            raise SkipTest('FTDI device is not a multi-port device')
        if width < 2:
            raise SkipTest('FTDI device does not support wide ports')
        url = url[:-1]
        cls.urls = [f'{url}1', f'{url}2']

    def test_default_gpio(self):
        """Check I/O.
        """
        gpio_in, gpio_out = GpioMpsseController(), GpioMpsseController()
        gpio_in.configure(self.urls[0], direction=0x0000, frequency=10e6,
                          debug=self.debug_mpsse)
        gpio_out.configure(self.urls[1], direction=0xFFFF, frequency=10e6,
                           debug=self.debug_mpsse)
        for out in range(3, 0x10000, 29):
            gpio_out.write(out)
            outv = gpio_out.read()[0]
            inv = gpio_in.read()[0]
            # check inputs match outputs
            self.assertEqual(inv, out)
            # check level of outputs match the ones written
            self.assertEqual(outv, out)
        gpio_in.close()
        gpio_out.close()

    def test_peek_gpio(self):
        """Check I/O peeking
        """
        gpio_in, gpio_out = GpioMpsseController(), GpioMpsseController()
        gpio_in.configure(self.urls[0], direction=0xFF00, frequency=10e6,
                          debug=self.debug)
        gpio_out.configure(self.urls[1], direction=0x00FF, frequency=10e6,
                           debug=self.debug)
        for out in range(256):
            gpio_out.write(out)
            outv = gpio_out.read()[0]
            inv = gpio_in.read(peek=True)
            # check inputs match outputs
            self.assertEqual(inv, out)
            #print(f'{out} {inv}')
            # check level of outputs match the ones written
            self.assertEqual(outv, out)
        gpio_in.close()
        gpio_out.close()

    def test_stream_gpio(self):
        """Check I/O streaming.

           Beware this test is CPU intensive w/ virtual framework
        """
        gpio_in, gpio_out = GpioMpsseController(), GpioMpsseController()
        gpio_in.configure(self.urls[0], direction=0x0000, frequency=10e6,
                          debug=self.debug)
        gpio_out.configure(self.urls[1], direction=0xFFFF, frequency=10e6,
                           debug=self.debug)
        outv = list(range(0, 0x10000, 29))
        max_count = min(gpio_out.ftdi.fifo_sizes[0],
                        gpio_in.ftdi.fifo_sizes[1])//2  # 2 bytes/value
        outv = outv[:max_count]
        gpio_out.write(outv)
        inv = gpio_in.read(len(outv))
        # for now, it is hard to test value exactness
        self.assertEqual(len(outv), len(inv))
        gpio_in.close()
        gpio_out.close()


def suite():
    suite_ = TestSuite()
    suite_.addTest(makeSuite(GpioAsyncTestCase, 'test'))
    suite_.addTest(makeSuite(GpioSyncTestCase, 'test'))
    suite_.addTest(makeSuite(GpioMpsseTestCase, 'test'))
    suite_.addTest(makeSuite(GpioMultiportTestCase, 'test'))
    return suite_


def virtualize():
    if not to_bool(environ.get('FTDI_VIRTUAL', 'off')):
        return
    from pyftdi.usbtools import UsbTools
    # Force PyUSB to use PyFtdi test framework for USB backends
    UsbTools.BACKENDS = ('backend.usbvirt', )
    # Ensure the virtual backend can be found and is loaded
    backend = UsbTools.find_backend()
    try:
        # obtain the loader class associated with the virtual backend
        global VirtLoader
        VirtLoader = backend.create_loader()
    except AttributeError as exc:
        raise AssertionError('Cannot load virtual USB backend') from exc


def main():
    import doctest
    doctest.testmod(modules[__name__])
    debug = to_bool(environ.get('FTDI_DEBUG', 'off'))
    if debug:
        formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(levelname)-7s'
                                      ' %(name)-20s [%(lineno)4d] %(message)s',
                                      '%H:%M:%S')
    else:
        formatter = logging.Formatter('%(message)s')
    level = environ.get('FTDI_LOGLEVEL', 'warning').upper()
    try:
        loglevel = getattr(logging, level)
    except AttributeError as exc:
        raise ValueError(f'Invalid log level: {level}') from exc
    FtdiLogger.log.addHandler(logging.StreamHandler(stdout))
    FtdiLogger.set_level(loglevel)
    FtdiLogger.set_formatter(formatter)
    virtualize()
    try:
        ut_main(defaultTest='suite')
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    # Useful environment variables:
    #  FTDI_DEVICE: a specific FTDI URL, default to ftdi:///1
    #  FTDI_LOGLEVEL: a Logger debug level, to define log verbosity
    #  FTDI_DEBUG: to enable/disable debug mode
    #  FTDI_VIRTUAL: to use a virtual device rather than a physical device
    main()



# Copyright (c) 2010-2020, Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2016, Emmanuel Bouaziz <ebouaziz@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

"""SPI support for PyFdti"""

#pylint: disable-msg=too-many-arguments
#pylint: disable-msg=too-many-locals
#pylint: disable-msg=too-many-branches
#pylint: disable-msg=too-many-statements
#pylint: disable-msg=too-many-instance-attributes
#pylint: disable-msg=too-many-public-methods
#pylint: disable-msg=invalid-name

from logging import getLogger
from struct import calcsize as scalc, pack as spack, unpack as sunpack
from threading import Lock
from typing import Any, Iterable, Mapping, Optional, Set, Union
from usb.core import Device as UsbDevice
from .ftdi import Ftdi, FtdiError


class SpiIOError(FtdiError):
    """SPI I/O error"""


class SpiPort:
    """SPI port

       An SPI port is never instanciated directly: use
       :py:meth:`SpiController.get_port()` method to obtain an SPI port.

       Example:

       >>> ctrl = SpiController()
       >>> ctrl.configure('ftdi://ftdi:232h/1')
       >>> spi = ctrl.get_port(1)
       >>> spi.set_frequency(1000000)
       >>> # send 2 bytes
       >>> spi.exchange([0x12, 0x34])
       >>> # send 2 bytes, then receive 2 bytes
       >>> out = spi.exchange([0x12, 0x34], 2)
       >>> # send 2 bytes, then receive 4 bytes, manage the transaction
       >>> out = spi.exchange([0x12, 0x34], 2, True, False)
       >>> out.extend(spi.exchange([], 2, False, True))
    """

    def __init__(self, controller: 'SpiController', cs: int, cs_hold: int = 3,
                 spi_mode: int = 0):
        self.log = getLogger('pyftdi.spi.port')
        self._controller = controller
        self._frequency = self._controller.frequency
        self._cs = cs
        self._cs_hold = cs_hold
        self.set_mode(spi_mode)

    def exchange(self, out: Union[bytes, bytearray, Iterable[int]] = b'',
                 readlen: int = 0, start: bool = True, stop: bool = True,
                 duplex: bool = False, droptail: int = 0) -> bytes:
        """Perform an exchange or a transaction with the SPI slave

           :param out: data to send to the SPI slave, may be empty to read out
                       data from the slave with no write.
           :param readlen: count of bytes to read out from the slave,
                       may be zero to only write to the slave
           :param start: whether to start an SPI transaction, i.e.
                        activate the /CS line for the slave. Use False to
                        resume a previously started transaction
           :param stop: whether to desactivete the /CS line for the slave.
                       Use False if the transaction should complete with a
                       further call to exchange()
           :param duplex: perform a full-duplex exchange (vs. half-duplex),
                          i.e. bits are clocked in and out at once.
           :param droptail: ignore up to 7 last bits (for non-byte sized SPI
                               accesses)
           :return: an array of bytes containing the data read out from the
                    slave
        """
        return self._controller.exchange(self._frequency, out, readlen,
                                         start and self._cs_prolog,
                                         stop and self._cs_epilog,
                                         self._cpol, self._cpha,
                                         duplex, droptail)

    def read(self, readlen: int = 0, start: bool = True, stop: bool = True,
             droptail: int = 0) -> bytes:
        """Read out bytes from the slave

           :param readlen: count of bytes to read out from the slave,
                           may be zero to only write to the slave
           :param start: whether to start an SPI transaction, i.e.
                        activate the /CS line for the slave. Use False to
                        resume a previously started transaction
           :param stop: whether to desactivete the /CS line for the slave.
                       Use False if the transaction should complete with a
                       further call to exchange()
           :param droptail: ignore up to 7 last bits (for non-byte sized SPI
                               accesses)
           :return: an array of bytes containing the data read out from the
                    slave
        """
        return self._controller.exchange(self._frequency, [], readlen,
                                         start and self._cs_prolog,
                                         stop and self._cs_epilog,
                                         self._cpol, self._cpha, False,
                                         droptail)

    def write(self, out: Union[bytes, bytearray, Iterable[int]],
              start: bool = True, stop: bool = True, droptail: int = 0) \
        -> None:
        """Write bytes to the slave

           :param out: data to send to the SPI slave, may be empty to read out
                       data from the slave with no write.
           :param start: whether to start an SPI transaction, i.e.
                        activate the /CS line for the slave. Use False to
                        resume a previously started transaction
           :param stop: whether to desactivete the /CS line for the slave.
                        Use False if the transaction should complete with a
                        further call to exchange()
           :param droptail: ignore up to 7 last bits (for non-byte sized SPI
                               accesses)
        """
        return self._controller.exchange(self._frequency, out, 0,
                                         start and self._cs_prolog,
                                         stop and self._cs_epilog,
                                         self._cpol, self._cpha, False,
                                         droptail)

    def flush(self) -> None:
        """Force the flush of the HW FIFOs"""
        self._controller.flush()

    def set_frequency(self, frequency: float):
        """Change SPI bus frequency

           :param float frequency: the new frequency in Hz
        """
        self._frequency = min(frequency, self._controller.frequency_max)

    def set_mode(self, mode: int, cs_hold: Optional[int] = None) -> None:
        """Set or change the SPI mode to communicate with the SPI slave.

           :param mode: new SPI mode
           :param cs_hold: change the /CS hold duration (or keep using previous
                           value)
        """
        if not 0 <= mode <= 3:
            raise SpiIOError('Invalid SPI mode: %d' % mode)
        if (mode & 0x2) and not self._controller.is_inverted_cpha_supported:
            raise SpiIOError('SPI with CPHA high is not supported by '
                             'this FTDI device')
        if cs_hold is None:
            cs_hold = self._cs_hold
        else:
            self._cs_hold = cs_hold
        self._cpol = bool(mode & 0x2)
        self._cpha = bool(mode & 0x1)
        cs_clock = 0xFF & ~((int(not self._cpol) and SpiController.SCK_BIT) |
                            SpiController.DO_BIT)
        cs_select = 0xFF & ~((SpiController.CS_BIT << self._cs) |
                             (int(not self._cpol) and SpiController.SCK_BIT) |
                             SpiController.DO_BIT)
        self._cs_prolog = bytes([cs_clock, cs_select])
        self._cs_epilog = bytes([cs_select] + [cs_clock] * int(cs_hold))

    def force_select(self, level: Optional[bool] = None,
                     cs_hold: float = 0) -> None:
        """Force-drive /CS signal.

           This API is not designed for a regular usage, but is reserved to
           very specific slave devices that require non-standard SPI
           signalling. There are very few use cases where this API is required.

           :param level: level to force on /CS output. This is a tri-state
                         value. A boolean value forces the selected signal
                         level; note that SpiPort no longer enforces that
                         following API calls generates valid SPI signalling:
                         use with extreme care. `None` triggers a pulse on /CS
                         output, i.e. /CS is not asserted once the method
                         returns, whatever the actual /CS level when this API
                         is called.
           :param cs_hold: /CS hold duration, as a unitless value. It is not
                         possible to control the exact duration of the pulse,
                         as it depends on the USB bus and the FTDI frequency.
        """
        clk, sel = self._cs_prolog
        if cs_hold:
            hold = max(1, cs_hold)
            if hold > SpiController.PAYLOAD_MAX_LENGTH:
                raise ValueError('cs_hold is too long')
        else:
            hold = self._cs_hold
        if level is None:
            seq = bytearray([clk])
            seq.extend([sel]*(1+hold))
            seq.extend([clk]*self._cs_hold)
        elif level:
            seq = bytearray([clk] * hold)
        else:
            seq = bytearray([clk] * hold)
            seq.extend([sel]*(1+hold))
        self._controller.force_control(self._frequency, bytes(seq))

    @property
    def frequency(self) -> float:
        """Return the current SPI bus block"""
        return self._frequency

    @property
    def cs(self) -> int:
        """Return the /CS index.

          :return: the /CS index (starting from 0)
        """
        return self._cs

    @property
    def mode(self) -> int:
        """Return the current SPI mode.

           :return: the SPI mode
        """
        return (int(self._cpol) << 2) | int(self._cpha)


class SpiGpioPort:
    """GPIO port

       A SpiGpioPort instance enables to drive GPIOs wich are not reserved for
       SPI feature as regular GPIOs.

       GPIO are managed as a bitfield. The LSBs are reserved for the SPI
       feature, which means that the lowest pin that can be used as a GPIO is
       *b4*:

       * *b0*: SPI SCLK
       * *b1*: SPI MOSI
       * *b2*: SPI MISO
       * *b3*: SPI CS0
       * *b4*: SPI CS1 or first GPIO

       If more than one SPI device is used, less GPIO pins are available, see
       the cs_count argument of the SpiController constructor.

       There is no offset bias in GPIO bit position, *i.e.* the first available
       GPIO can be reached from as ``0x10``.

       Bitfield size depends on the FTDI device: 4432H series use 8-bit GPIO
       ports, while 232H and 2232H series use wide 16-bit ports.

       An SpiGpio port is never instanciated directly: use
       :py:meth:`SpiController.get_gpio()` method to obtain the GPIO port.
    """
    def __init__(self, controller: 'SpiController'):
        self.log = getLogger('pyftdi.spi.gpio')
        self._controller = controller

    @property
    def pins(self) -> int:
        """Report the configured GPIOs as a bitfield.

           A true bit represents a GPIO, a false bit a reserved or not
           configured pin.

           :return: the bitfield of configured GPIO pins.
        """
        return self._controller.gpio_pins

    @property
    def all_pins(self) -> int:
        """Report the addressable GPIOs as a bitfield.

           A true bit represents a pin which may be used as a GPIO, a false bit
           a reserved pin (for SPI support)

           :return: the bitfield of configurable GPIO pins.
        """
        return self._controller.gpio_all_pins

    @property
    def width(self) -> int:
        """Report the FTDI count of addressable pins.

           Note that all pins, including reserved SPI ones, are reported.

           :return: the count of IO pins (including SPI ones).
        """
        return self._controller.width

    @property
    def direction(self) -> int:
        """Provide the FTDI GPIO direction.self

           A true bit represents an output GPIO, a false bit an input GPIO.

           :return: the bitfield of direction.
        """
        return self._controller.direction

    def read(self, with_output: bool = False) -> int:
        """Read GPIO port.

           :param with_output: set to unmask output pins
           :return: the GPIO port pins as a bitfield
        """
        return self._controller.read_gpio(with_output)

    def write(self, value: int) -> None:
        """Write GPIO port.

           :param value: the GPIO port pins as a bitfield
        """
        return self._controller.write_gpio(value)

    def set_direction(self, pins: int, direction: int) -> None:
        """Change the direction of the GPIO pins.

           :param pins: which GPIO pins should be reconfigured
           :param direction: direction bitfield (high level for output)
        """
        self._controller.set_gpio_direction(pins, direction)


class SpiController:
    """SPI master.

        :param int cs_count: is the number of /CS lines (one per device to
            drive on the SPI bus)
        :param turbo: increase throughput over USB bus, but may not be
                      supported with some specific slaves
    """

    SCK_BIT = 0x01
    DO_BIT = 0x02
    DI_BIT = 0x04
    CS_BIT = 0x08
    SPI_BITS = DI_BIT | DO_BIT | SCK_BIT
    PAYLOAD_MAX_LENGTH = 0xFF00  # 16 bits max (- spare for control)

    def __init__(self, cs_count: int = 1, turbo: bool = True):
        self.log = getLogger('pyftdi.spi.ctrl')
        self._ftdi = Ftdi()
        self._lock = Lock()
        self._gpio_port = None
        self._gpio_dir = 0
        self._gpio_mask = 0
        self._gpio_low = 0
        self._wide_port = False
        self._cs_count = cs_count
        self._turbo = turbo
        self._immediate = bytes((Ftdi.SEND_IMMEDIATE,))
        self._frequency = 0.0
        self._clock_phase = False
        self._cs_bits = 0
        self._spi_ports = []
        self._spi_dir = 0
        self._spi_mask = self.SPI_BITS

    def configure(self, url: Union[str, UsbDevice],
                  **kwargs: Mapping[str, Any]) -> None:
        """Configure the FTDI interface as a SPI master

           :param url: FTDI URL string, such as ``ftdi://ftdi:232h/1``
           :param kwargs: options to configure the SPI bus

           Accepted options:

           * ``interface``: when URL is specifed as a USB device, the interface
             named argument can be used to select a specific port of the FTDI
             device, as an integer starting from 1.
           * ``direction`` a bitfield specifying the FTDI GPIO direction,
             where high level defines an output, and low level defines an
             input. Only useful to setup default IOs at start up, use
             :py:class:`SpiGpioPort` to drive GPIOs. Note that pins reserved
             for SPI feature take precedence over any this setting.
           * ``initial`` a bitfield specifying the initial output value. Only
             useful to setup default IOs at start up, use
             :py:class:`SpiGpioPort` to drive GPIOs.
           * ``frequency`` the SPI bus frequency in Hz. Note that each slave
             may reconfigure the SPI bus with a specialized
             frequency.
           * ``cs_count`` count of chip select signals dedicated to select
             SPI slave devices, starting from A*BUS3 pin
           * ``turbo`` whether to enable or disable turbo mode
           * ``debug`` to increase log verbosity, using MPSSE tracer
        """
        # it is better to specify CS and turbo in configure, but the older
        # API where these parameters are specified at instanciation has been
        # preserved
        if 'cs_count' in kwargs:
            self._cs_count = int(kwargs['cs_count'])
            del kwargs['cs_count']
        if not 1 <= self._cs_count <= 5:
            raise ValueError('Unsupported CS line count: %d' %
                             self._cs_count)
        if 'turbo' in kwargs:
            self._turbo = bool(kwargs['turbo'])
            del kwargs['turbo']
        if 'direction' in kwargs:
            io_dir = int(kwargs['direction'])
            del kwargs['direction']
        else:
            io_dir = 0
        if 'initial' in kwargs:
            io_out = int(kwargs['initial'])
            del kwargs['initial']
        else:
            io_out = 0
        if 'interface' in kwargs:
            if isinstance(url, str):
                raise SpiIOError('url and interface are mutually exclusive')
            interface = int(kwargs['interface'])
            del kwargs['interface']
        else:
            interface = 1
        with self._lock:
            if self._frequency > 0.0:
                raise SpiIOError('Already configured')
            self._cs_bits = (((SpiController.CS_BIT << self._cs_count) - 1) &
                             ~(SpiController.CS_BIT - 1))
            self._spi_ports = [None] * self._cs_count
            self._spi_dir = (self._cs_bits |
                             SpiController.DO_BIT |
                             SpiController.SCK_BIT)
            self._spi_mask = self._cs_bits | self.SPI_BITS
            # until the device is open, there is no way to tell if it has a
            # wide (16) or narrow port (8). Lower API can deal with any, so
            # delay any truncation till the device is actually open
            self._set_gpio_direction(16, (~self._spi_mask) & 0xFFFF, io_dir)
            kwargs['direction'] = self._spi_dir | self._gpio_dir
            kwargs['initial'] = self._cs_bits | (io_out & self._gpio_mask)
            if not isinstance(url, str):
                self._frequency = self._ftdi.open_mpsse_from_device(
                    url, interface=interface, **kwargs)
            else:
                self._frequency = self._ftdi.open_mpsse_from_url(url, **kwargs)
            self._ftdi.enable_adaptive_clock(False)
            self._wide_port = self._ftdi.has_wide_port
            if not self._wide_port:
                self._set_gpio_direction(8, io_out & 0xFF, io_dir & 0xFF)

    def close(self, freeze: bool = False) -> None:
        """Close the FTDI interface.

           :param freeze: if set, FTDI port is not reset to its default
                          state on close.
        """
        with self._lock:
            if self._ftdi.is_connected:
                self._ftdi.close(freeze)
        self._frequency = 0.0

    def terminate(self) -> None:
        """Close the FTDI interface.

           :note: deprecated API, use close()
        """
        self.close()

    def get_port(self, cs: int, freq: Optional[float] = None,
                 mode: int = 0) -> SpiPort:
        """Obtain a SPI port to drive a SPI device selected by Chip Select.

           :note: SPI mode 1 and 3 are not officially supported.

           :param cs: chip select slot, starting from 0
           :param freq: SPI bus frequency for this slave in Hz
           :param mode: SPI mode [0, 1, 2, 3]
        """
        with self._lock:
            if not self._ftdi.is_connected:
                raise SpiIOError("FTDI controller not initialized")
            if cs >= len(self._spi_ports):
                if cs < 5:
                    # increase cs_count (up to 4) to reserve more /CS channels
                    raise SpiIOError("/CS pin %d not reserved for SPI" % cs)
                raise SpiIOError("No such SPI port: %d" % cs)
            if not self._spi_ports[cs]:
                freq = min(freq or self._frequency, self.frequency_max)
                hold = freq and (1+int(1E6/freq))
                self._spi_ports[cs] = SpiPort(self, cs, cs_hold=hold,
                                              spi_mode=mode)
                self._spi_ports[cs].set_frequency(freq)
                self._flush()
            return self._spi_ports[cs]

    def get_gpio(self) -> SpiGpioPort:
        """Retrieve the GPIO port.

           :return: GPIO port
        """
        with self._lock:
            if not self._ftdi.is_connected:
                raise SpiIOError("FTDI controller not initialized")
            if not self._gpio_port:
                self._gpio_port = SpiGpioPort(self)
            return self._gpio_port

    @property
    def ftdi(self) -> Ftdi:
        """Return the Ftdi instance.

           :return: the Ftdi instance
        """
        return self._ftdi

    @property
    def configured(self) -> bool:
        """Test whether the device has been properly configured.

           :return: True if configured
        """
        return self._ftdi.is_connected

    @property
    def frequency_max(self) -> float:
        """Provides the maximum SPI clock frequency in Hz.

           :return: SPI bus clock frequency
        """
        return self._ftdi.frequency_max

    @property
    def frequency(self) -> float:
        """Provides the current SPI clock frequency in Hz.

           :return: the SPI bus clock frequency
        """
        return self._frequency

    @property
    def direction(self):
        """Provide the FTDI pin direction

           A true bit represents an output pin, a false bit an input pin.

           :return: the bitfield of direction.
        """
        return self._spi_dir | self._gpio_dir

    @property
    def channels(self) -> int:
        """Provide the maximum count of slaves.


           :return: the count of pins reserved to drive the /CS signal
        """
        return self._cs_count

    @property
    def active_channels(self) -> Set[int]:
        """Provide the set of configured slaves /CS.

           :return: Set of /CS, one for each configured slaves
        """
        return {port[0] for port in enumerate(self._spi_ports) if port[1]}

    @property
    def gpio_pins(self):
        """Report the configured GPIOs as a bitfield.

           A true bit represents a GPIO, a false bit a reserved or not
           configured pin.

           :return: the bitfield of configured GPIO pins.
        """
        with self._lock:
            return self._gpio_mask

    @property
    def gpio_all_pins(self):
        """Report the addressable GPIOs as a bitfield.

           A true bit represents a pin which may be used as a GPIO, a false bit
           a reserved pin (for SPI support)

           :return: the bitfield of configurable GPIO pins.
        """
        mask = (1 << self.width) - 1
        with self._lock:
            return mask & ~self._spi_mask

    @property
    def width(self):
        """Report the FTDI count of addressable pins.

           :return: the count of IO pins (including SPI ones).
        """
        return 16 if self._wide_port else 8

    @property
    def is_inverted_cpha_supported(self) -> bool:
        """Report whether it is possible to supported CPHA=1.

           :return: inverted CPHA supported (with a kludge)
        """
        return self._ftdi.is_H_series

    def exchange(self, frequency: float,
                 out: Union[bytes, bytearray, Iterable[int]], readlen: int,
                 cs_prolog: Optional[bytes] = None,
                 cs_epilog: Optional[bytes] = None,
                 cpol: bool = False, cpha: bool = False,
                 duplex: bool = False, droptail: int = 0) -> bytes:
        """Perform an exchange or a transaction with the SPI slave

           :param out: data to send to the SPI slave, may be empty to read out
                       data from the slave with no write.
           :param readlen: count of bytes to read out from the slave,
                           may be zero to only write to the slave,
           :param cs_prolog: the prolog MPSSE command sequence to execute
                             before the actual exchange.
           :param cs_epilog: the epilog MPSSE command sequence to execute
                             after the actual exchange.
           :param cpol: SPI clock polarity, derived from the SPI mode
           :param cpol: SPI clock phase, derived from the SPI mode
           :param duplex: perform a full-duplex exchange (vs. half-duplex),
                          i.e. bits are clocked in and out at once or
                          in a write-then-read manner.
           :param droptail: ignore up to 7 last bits (for non-byte sized SPI
                             accesses)
           :return: bytes containing the data read out from the slave, if any
        """
        if not 0 <= droptail <= 7:
            raise ValueError('Invalid skip bit count')
        if duplex:
            if readlen > len(out):
                tmp = bytearray(out)
                tmp.extend([0] * (readlen - len(out)))
                out = tmp
            elif not readlen:
                readlen = len(out)
        with self._lock:
            if duplex:
                data = self._exchange_full_duplex(frequency, out,
                                                  cs_prolog, cs_epilog,
                                                  cpol, cpha, droptail)
                return data[:readlen]
            return self._exchange_half_duplex(frequency, out, readlen,
                                              cs_prolog, cs_epilog,
                                              cpol, cpha, droptail)

    def force_control(self, frequency: float, sequence: bytes) -> None:
        """Execution an arbitrary SPI control bit sequence.
           Use with extreme care, as it may lead to unexpected results. Regular
           usage of SPI does not require to invoke this API.

           :param sequence: the bit sequence to execute.
        """
        with self._lock:
            self._force(frequency, sequence)

    def flush(self) -> None:
        """Flush the HW FIFOs.
        """
        with self._lock:
            self._flush()

    def read_gpio(self, with_output: bool = False) -> int:
        """Read GPIO port

           :param  with_output: set to unmask output pins
           :return: the GPIO port pins as a bitfield
        """
        with self._lock:
            data = self._read_raw(self._wide_port)
        value = data & self._gpio_mask
        if not with_output:
            value &= ~self._gpio_dir
        return value

    def write_gpio(self, value: int) -> None:
        """Write GPIO port

           :param value: the GPIO port pins as a bitfield
        """
        with self._lock:
            if (value & self._gpio_dir) != value:
                raise SpiIOError('No such GPO pins: %04x/%04x' %
                                 (self._gpio_dir, value))
            # perform read-modify-write
            use_high = self._wide_port and (self.direction & 0xff00)
            data = self._read_raw(use_high)
            data &= ~self._gpio_mask
            data |= value
            self._write_raw(data, use_high)
            self._gpio_low = data & 0xFF & ~self._spi_mask

    def set_gpio_direction(self, pins: int, direction: int) -> None:
        """Change the direction of the GPIO pins

           :param pins: which GPIO pins should be reconfigured
           :param direction: direction bitfield (on for output)
        """
        with self._lock:
            self._set_gpio_direction(16 if self._wide_port else 8,
                                     pins, direction)

    def _set_gpio_direction(self, width: int, pins: int,
                            direction: int) -> None:
        if pins & self._spi_mask:
            raise SpiIOError('Cannot access SPI pins as GPIO')
        gpio_mask = (1 << width) - 1
        gpio_mask &= ~self._spi_mask
        if (pins & gpio_mask) != pins:
            raise SpiIOError('No such GPIO pin(s)')
        self._gpio_dir &= ~pins
        self._gpio_dir |= (pins & direction)
        self._gpio_mask = gpio_mask & pins

    def _read_raw(self, read_high: bool) -> int:
        if not self._ftdi.is_connected:
            raise SpiIOError("FTDI controller not initialized")
        if read_high:
            cmd = bytes([Ftdi.GET_BITS_LOW,
                         Ftdi.GET_BITS_HIGH,
                         Ftdi.SEND_IMMEDIATE])
            fmt = '<H'
        else:
            cmd = bytes([Ftdi.GET_BITS_LOW,
                         Ftdi.SEND_IMMEDIATE])
            fmt = 'B'
        self._ftdi.write_data(cmd)
        size = scalc(fmt)
        data = self._ftdi.read_data_bytes(size, 4)
        if len(data) != size:
            raise SpiIOError('Cannot read GPIO')
        value, = sunpack(fmt, data)
        return value

    def _write_raw(self, data: int, write_high: bool) -> None:
        if not self._ftdi.is_connected:
            raise SpiIOError("FTDI controller not initialized")
        direction = self.direction
        low_data = data & 0xFF
        low_dir = direction & 0xFF
        if write_high:
            high_data = (data >> 8) & 0xFF
            high_dir = (direction >> 8) & 0xFF
            cmd = bytes([Ftdi.SET_BITS_LOW, low_data, low_dir,
                         Ftdi.SET_BITS_HIGH, high_data, high_dir])
        else:
            cmd = bytes([Ftdi.SET_BITS_LOW, low_data, low_dir])
        self._ftdi.write_data(cmd)

    def _force(self, frequency: float, sequence: bytes):
        if not self._ftdi.is_connected:
            raise SpiIOError("FTDI controller not initialized")
        if len(sequence) > SpiController.PAYLOAD_MAX_LENGTH:
            raise SpiIOError("Output payload is too large")
        if self._frequency != frequency:
            self._ftdi.set_frequency(frequency)
            # store the requested value, not the actual one (best effort),
            # to avoid setting unavailable values on each call.
            self._frequency = frequency
        cmd = bytearray()
        direction = self.direction & 0xFF
        for ctrl in sequence:
            ctrl &= self._spi_mask
            ctrl |= self._gpio_low
            cmd.extend((Ftdi.SET_BITS_LOW, ctrl, direction))
        self._ftdi.write_data(cmd)

    def _exchange_half_duplex(self, frequency: float,
                              out: Union[bytes, bytearray, Iterable[int]],
                              readlen: int, cs_prolog: bytes, cs_epilog: bytes,
                              cpol: bool, cpha: bool,
                              droptail: int) -> bytes:
        if not self._ftdi.is_connected:
            raise SpiIOError("FTDI controller not initialized")
        if len(out) > SpiController.PAYLOAD_MAX_LENGTH:
            raise SpiIOError("Output payload is too large")
        if readlen > SpiController.PAYLOAD_MAX_LENGTH:
            raise SpiIOError("Input payload is too large")
        if cpha:
            # to enable CPHA, we need to use a workaround with FTDI device,
            # that is enable 3-phase clocking (which is usually dedicated to
            # I2C support). This mode use use 3 clock period instead of 2,
            # which implies the FTDI frequency should be fixed to match the
            # requested one.
            frequency = (3*frequency)//2
        if self._frequency != frequency:
            self._ftdi.set_frequency(frequency)
            # store the requested value, not the actual one (best effort),
            # to avoid setting unavailable values on each call.
            self._frequency = frequency
        direction = self.direction & 0xFF  # low bits only
        cmd = bytearray()
        for ctrl in cs_prolog or []:
            ctrl &= self._spi_mask
            ctrl |= self._gpio_low
            cmd.extend((Ftdi.SET_BITS_LOW, ctrl, direction))
        epilog = bytearray()
        if cs_epilog:
            for ctrl in cs_epilog:
                ctrl &= self._spi_mask
                ctrl |= self._gpio_low
                epilog.extend((Ftdi.SET_BITS_LOW, ctrl, direction))
            # Restore idle state
            cs_high = [Ftdi.SET_BITS_LOW, self._cs_bits | self._gpio_low,
                       direction]
            if not self._turbo:
                cs_high.append(Ftdi.SEND_IMMEDIATE)
            epilog.extend(cs_high)
        writelen = len(out)
        if self._clock_phase != cpha:
            self._ftdi.enable_3phase_clock(cpha)
            self._clock_phase = cpha
        if writelen:
            if not droptail:
                wcmd = (Ftdi.WRITE_BYTES_NVE_MSB if not cpol else
                        Ftdi.WRITE_BYTES_PVE_MSB)
                write_cmd = spack('<BH', wcmd, writelen-1)
                cmd.extend(write_cmd)
                cmd.extend(out)
            else:
                bytelen = writelen-1
                if bytelen:
                    wcmd = (Ftdi.WRITE_BYTES_NVE_MSB if not cpol else
                            Ftdi.WRITE_BYTES_PVE_MSB)
                    write_cmd = spack('<BH', wcmd, bytelen-1)
                    cmd.extend(write_cmd)
                    cmd.extend(out[:-1])
                wcmd = (Ftdi.WRITE_BITS_NVE_MSB if not cpol else
                        Ftdi.WRITE_BITS_PVE_MSB)
                write_cmd = spack('<BBB', wcmd, 7-droptail, out[-1])
                cmd.extend(write_cmd)
        if readlen:
            if not droptail:
                rcmd = (Ftdi.READ_BYTES_NVE_MSB if not cpol else
                        Ftdi.READ_BYTES_PVE_MSB)
                read_cmd = spack('<BH', rcmd, readlen-1)
                cmd.extend(read_cmd)
            else:
                bytelen = readlen-1
                if bytelen:
                    rcmd = (Ftdi.READ_BYTES_NVE_MSB if not cpol else
                            Ftdi.READ_BYTES_PVE_MSB)
                    read_cmd = spack('<BH', rcmd, bytelen-1)
                    cmd.extend(read_cmd)
                rcmd = (Ftdi.READ_BITS_NVE_MSB if not cpol else
                        Ftdi.READ_BITS_PVE_MSB)
                read_cmd = spack('<BB', rcmd, 7-droptail)
                cmd.extend(read_cmd)
            cmd.extend(self._immediate)
            if self._turbo:
                if epilog:
                    cmd.extend(epilog)
                self._ftdi.write_data(cmd)
            else:
                self._ftdi.write_data(cmd)
                if epilog:
                    self._ftdi.write_data(epilog)
            # USB read cycle may occur before the FTDI device has actually
            # sent the data, so try to read more than once if no data is
            # actually received
            data = self._ftdi.read_data_bytes(readlen, 4)
            if droptail:
                data[-1] = 0xff & (data[-1] << droptail)
        else:
            if writelen:
                if self._turbo:
                    if epilog:
                        cmd.extend(epilog)
                    self._ftdi.write_data(cmd)
                else:
                    self._ftdi.write_data(cmd)
                    if epilog:
                        self._ftdi.write_data(epilog)
            data = bytearray()
        return data

    def _exchange_full_duplex(self, frequency: float,
                              out: Union[bytes, bytearray, Iterable[int]],
                              cs_prolog: bytes, cs_epilog: bytes,
                              cpol: bool, cpha: bool,
                              droptail: int) -> bytes:
        if not self._ftdi.is_connected:
            raise SpiIOError("FTDI controller not initialized")
        if len(out) > SpiController.PAYLOAD_MAX_LENGTH:
            raise SpiIOError("Output payload is too large")
        if cpha:
            # to enable CPHA, we need to use a workaround with FTDI device,
            # that is enable 3-phase clocking (which is usually dedicated to
            # I2C support). This mode use use 3 clock period instead of 2,
            # which implies the FTDI frequency should be fixed to match the
            # requested one.
            frequency = (3*frequency)//2
        if self._frequency != frequency:
            self._ftdi.set_frequency(frequency)
            # store the requested value, not the actual one (best effort),
            # to avoid setting unavailable values on each call.
            self._frequency = frequency
        direction = self.direction & 0xFF  # low bits only
        cmd = bytearray()
        for ctrl in cs_prolog or []:
            ctrl &= self._spi_mask
            ctrl |= self._gpio_low
            cmd.extend((Ftdi.SET_BITS_LOW, ctrl, direction))
        epilog = bytearray()
        if cs_epilog:
            for ctrl in cs_epilog:
                ctrl &= self._spi_mask
                ctrl |= self._gpio_low
                epilog.extend((Ftdi.SET_BITS_LOW, ctrl, direction))
            # Restore idle state
            cs_high = [Ftdi.SET_BITS_LOW, self._cs_bits | self._gpio_low,
                       direction]
            if not self._turbo:
                cs_high.append(Ftdi.SEND_IMMEDIATE)
            epilog.extend(cs_high)
        exlen = len(out)
        if self._clock_phase != cpha:
            self._ftdi.enable_3phase_clock(cpha)
            self._clock_phase = cpha
        if not droptail:
            wcmd = (Ftdi.RW_BYTES_PVE_NVE_MSB if not cpol else
                    Ftdi.RW_BYTES_NVE_PVE_MSB)
            write_cmd = spack('<BH', wcmd, exlen-1)
            cmd.extend(write_cmd)
            cmd.extend(out)
        else:
            bytelen = exlen-1
            if bytelen:
                wcmd = (Ftdi.RW_BYTES_PVE_NVE_MSB if not cpol else
                        Ftdi.RW_BYTES_NVE_PVE_MSB)
                write_cmd = spack('<BH', wcmd, bytelen-1)
                cmd.extend(write_cmd)
                cmd.extend(out[:-1])
            wcmd = (Ftdi.RW_BITS_PVE_NVE_MSB if not cpol else
                    Ftdi.RW_BITS_NVE_PVE_MSB)
            write_cmd = spack('<BBB', wcmd, 7-droptail, out[-1])
            cmd.extend(write_cmd)
        cmd.extend(self._immediate)
        if self._turbo:
            if epilog:
                cmd.extend(epilog)
            self._ftdi.write_data(cmd)
        else:
            self._ftdi.write_data(cmd)
            if epilog:
                self._ftdi.write_data(epilog)
        # USB read cycle may occur before the FTDI device has actually
        # sent the data, so try to read more than once if no data is
        # actually received
        data = self._ftdi.read_data_bytes(exlen, 4)
        if droptail:
            data[-1] = 0xff & (data[-1] << droptail)
        return data

    def _flush(self) -> None:
        self._ftdi.write_data(self._immediate)
        self._ftdi.purge_buffers()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2010-2020, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import logging
from doctest import testmod
from os import environ
from sys import modules, stdout
from time import sleep, time as now
from unittest import TestCase, TestSuite, SkipTest, makeSuite, main as ut_main
from pyftdi import FtdiLogger
from pyftdi.ftdi import Ftdi, FtdiError
from pyftdi.usbtools import UsbTools, UsbToolsError


class FtdiTestCase(TestCase):
    """FTDI driver test case"""

    def test_multiple_interface(self):
        # the following calls used to create issues (several interfaces from
        # the same device). The test expects an FTDI 2232H here
        ftdi1 = Ftdi()
        ftdi1.open(vendor=0x403, product=0x6010, interface=1)
        ftdi2 = Ftdi()
        ftdi2.open(vendor=0x403, product=0x6010, interface=2)
        for _ in range(5):
            print("If#1: ", hex(ftdi1.poll_modem_status()))
            print("If#2: ", ftdi2.modem_status())
            sleep(0.500)
        ftdi1.close()
        ftdi2.close()


class HotplugTestCase(TestCase):

    def test_hotplug_discovery(self):
        """Demonstrate how to connect to an hotplugged FTDI device, i.e.
           an FTDI device that is connected after the initial attempt to
           enumerate it on the USB bus."""
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        ftdi = Ftdi()
        timeout = now() + 5.0  # sanity check: bail out after 10 seconds
        while now() < timeout:
            try:
                ftdi.open_from_url(url)
                break
            except UsbToolsError:
                UsbTools.flush_cache()
                sleep(0.05)
                continue
        self.assertTrue(ftdi.is_connected, 'Unable to connect to FTDI')
        print('Connected to FTDI', url)


class ResetTestCase(TestCase):

    def test_simple_reset(self):
        """Demonstrate how to connect to an hotplugged FTDI device, i.e.
           an FTDI device that is connected after the initial attempt to
           enumerate it on the USB bus."""
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        ftdi = Ftdi()
        ftdi.open_from_url(url)
        self.assertTrue(ftdi.is_connected, 'Unable to connect to FTDI')
        ftdi.close()
        self.assertFalse(ftdi.is_connected, 'Unable to close connection')
        ftdi.open_from_url(url)
        self.assertTrue(ftdi.is_connected, 'Unable to connect to FTDI')
        ftdi.reset(False)

    def test_dual_if_reset(self):
        """Demonstrate how to connect to an hotplugged FTDI device, i.e.
           an FTDI device that is connected after the initial attempt to
           enumerate it on the USB bus."""
        url1 = environ.get('FTDI_DEVICE', 'ftdi:///1')
        ftdi1 = Ftdi()
        ftdi1.open_from_url(url1)
        count = ftdi1.device_port_count
        if count < 2:
            ftdi1.close()
            raise SkipTest('FTDI device is not a multi-port device')
        next_port = (int(url1[-1]) % count) + 1
        url2 = 'ftdi:///%d' % next_port
        ftdi2 = Ftdi()
        self.assertTrue(ftdi1.is_connected, 'Unable to connect to FTDI')
        ftdi2.open_from_url(url2)
        # use latenty setting to set/test configuration is preserved
        ftdi2.set_latency_timer(128)
        # should be the same value
        self.assertEqual(ftdi2.get_latency_timer(), 128)
        self.assertTrue(ftdi2.is_connected, 'Unable to connect to FTDI')
        ftdi1.close()
        self.assertFalse(ftdi1.is_connected, 'Unable to close connection')
        # closing first connection should not alter second interface
        self.assertEqual(ftdi2.get_latency_timer(), 128)
        ftdi1.open_from_url(url1)
        self.assertTrue(ftdi1.is_connected, 'Unable to connect to FTDI')
        # a FTDI reset should not alter settings...
        ftdi1.reset(False)
        self.assertEqual(ftdi2.get_latency_timer(), 128)
        # ... however performing a USB reset through any interface should alter
        # any previous settings made to all interfaces
        ftdi1.reset(True)
        self.assertNotEqual(ftdi2.get_latency_timer(), 128)


class DisconnectTestCase(TestCase):
    """This test requires user interaction to unplug/plug back the device.
    """

    def test_close_on_disconnect(self):
        """Validate close after disconnect."""
        log = logging.getLogger('pyftdi.tests.ftdi')
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        ftdi = Ftdi()
        ftdi.open_from_url(url)
        self.assertTrue(ftdi.is_connected, 'Unable to connect to FTDI')
        print('Please disconnect FTDI device')
        while ftdi.is_connected:
            try:
                ftdi.poll_modem_status()
            except FtdiError:
                break
            sleep(0.1)
        ftdi.close()
        print('Please reconnect FTDI device')
        while True:
            UsbTools.flush_cache()
            try:
                ftdi.open_from_url(url)
            except (FtdiError, UsbToolsError):
                log.debug('FTDI device not detected')
                sleep(0.1)
            except ValueError:
                log.warning('FTDI device not initialized')
                ftdi.close()
                sleep(0.1)
            else:
                log.info('FTDI device detected')
                break
        ftdi.poll_modem_status()
        ftdi.close()


def suite():
    suite_ = TestSuite()
    #suite_.addTest(makeSuite(FtdiTestCase, 'test'))
    #suite_.addTest(makeSuite(HotplugTestCase, 'test'))
    suite_.addTest(makeSuite(ResetTestCase, 'test'))
    suite_.addTest(makeSuite(DisconnectTestCase, 'test'))
    return suite_


if __name__ == '__main__':
    testmod(modules[__name__])
    FtdiLogger.log.addHandler(logging.StreamHandler(stdout))
    level = environ.get('FTDI_LOGLEVEL', 'info').upper()
    try:
        loglevel = getattr(logging, level)
    except AttributeError as exc:
        raise ValueError(f'Invalid log level: {level}') from exc
    FtdiLogger.set_level(loglevel)
    ut_main(defaultTest='suite')



"""Constant importer from existing modules."""

# Copyright (c) 2020-2021, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

#pylint: disable-msg=missing-docstring
#pylint: disable-msg=invalid-name
#pylint: disable-msg=too-many-instance-attributes

from enum import Enum
from importlib import import_module
from sys import version_info
from pyftdi.ftdi import Ftdi
from pyftdi.misc import EasyDict


class UsbConstants:
    """Expose useful constants defined in PyUSB and allow reverse search, i.e.
       retrieve constant literals from integral values.
    """

    DEVICE_REQUESTS = {
        (True, 0x0): 'get_status',
        (False, 0x1): 'clear_feature',
        (False, 0x3): 'set_feature',
        (False, 0x5): 'set_address',
        (True, 0x6): 'get_descriptor',
        (False, 0x7): 'set_descriptor',
        (True, 0x8): 'get_configuration',
        (False, 0x9): 'set_configuration',
    }

    def __init__(self):
        self._desc_type = self._load_constants('desc_type')
        self._desc_type_mask = self._mask(self._desc_type)
        self._ctrl_dir = self._load_constants('ctrl')
        self._ctrl_dir_mask = self._mask(self._ctrl_dir)
        self._ctrl_type = self._load_constants('ctrl_type')
        self._ctrl_type_mask = self._mask(self._ctrl_type)
        self._ctrl_recipient = self._load_constants('ctrl_recipient')
        self._ctrl_recipient_mask = self._mask(self._ctrl_recipient)
        self._endpoint_type = self._load_constants('endpoint_type')
        self._endpoint_type_mask = self._mask(self._endpoint_type)
        self._descriptors = EasyDict({v.upper(): k
                                      for k, v in self._desc_type.items()})
        self.endpoints = self._load_constants('endpoint', True)
        self.endpoint_types = self._load_constants('endpoint_type', True)
        self.speeds = self._load_constants('speed', True)

    @property
    def descriptors(self):
        return self._descriptors

    @classmethod
    def _load_constants(cls, prefix: str, reverse=False):
        prefix = prefix.upper()
        if not prefix.endswith('_'):
            prefix = f'{prefix}_'
        mod = import_module('usb.util')
        mapping = EasyDict()
        plen = len(prefix)
        for entry in dir(mod):
            if not entry.startswith(prefix):
                continue
            if '_' in entry[plen:]:
                continue
            if not reverse:
                mapping[getattr(mod, entry)] = entry[plen:].lower()
            else:
                mapping[entry[plen:].lower()] = getattr(mod, entry)
        if not mapping:
            raise ValueError(f"No USB constant found for {prefix.rstrip('_')}")
        return mapping

    @classmethod
    def _mask(cls, mapping: dict) -> int:
        mask = 0
        for val in mapping:
            mask |= val
        return mask

    def is_req_out(self, reqtype: int) -> str:
        return not reqtype & self._ctrl_dir_mask

    def dec_req_ctrl(self, reqtype: int) -> str:
        return self._ctrl_dir[reqtype & self._ctrl_dir_mask]

    def dec_req_type(self, reqtype: int) -> str:
        return self._ctrl_type[reqtype & self._ctrl_type_mask]

    def dec_req_rcpt(self, reqtype: int) -> str:
        return self._ctrl_recipient[reqtype & self._ctrl_recipient_mask]

    def dec_req_name(self, reqtype: int, request: int) -> str:
        direction = bool(reqtype & self._ctrl_dir_mask)
        try:
            return self.DEVICE_REQUESTS[(direction, request)]
        except KeyError:
            return f'req x{request:02x}'

    def dec_desc_type(self, desctype: int) -> str:
        return self._desc_type[desctype & self._desc_type_mask]



class FtdiConstants:
    """Expose useful constants defined in Ftdi and allow reverse search, i.e.
       retrieve constant literals from integral values.
    """

    def __init__(self):
        self._dcache = {}
        self._rcache = {}

    @classmethod
    def _load_constants(cls, prefix: str, reverse=False):
        prefix = prefix.upper()
        if not prefix.endswith('_'):
            prefix = f'{prefix}_'
        mapping = EasyDict()
        plen = len(prefix)
        for name in dir(Ftdi):
            if not name.startswith(prefix):
                continue
            if not reverse:
                mapping[getattr(Ftdi, name)] = name[plen:].lower()
            else:
                mapping[name[plen:].lower()] = getattr(Ftdi, name)
        if not mapping:
            # maybe an enum
            prefix = prefix.rstrip('_').lower()
            for name in dir(Ftdi):
                if not name.lower().startswith(prefix):
                    continue
                item = getattr(Ftdi, name)
                if issubclass(item, Enum):
                    if not reverse:
                        mapping = {en.value: en.name.lower() for en in item}
                    else:
                        mapping = {en.name.lower(): en.value for en in item}
        if not mapping:
            raise ValueError(f"No FTDI constant found for "
                             f"{prefix.rstrip('_')}")
        return mapping

    def get_name(self, prefix: str, value: int) -> str:
        if prefix not in self._dcache:
            self._dcache[prefix] = self._load_constants(prefix)
        try:
            return self._dcache[prefix][value]
        except KeyError:
            return f'x?{value:04x}'

    def get_value(self, prefix: str, name: str) -> str:
        if prefix not in self._rcache:
            self._rcache[prefix] = self._load_constants(prefix, True)
        try:
            return self._rcache[prefix][name.lower()]
        except KeyError as exc:
            raise ValueError(f'Unknown name {prefix}.{name}') from exc

    def dec_req_name(self, request: int) -> str:
        return self.get_name('sio_req', request)


USBCONST = UsbConstants()
"""Unique instance of USB constant container."""

FTDICONST = FtdiConstants()
"""Unique instances of FTDI constant container."""



#!/usr/bin/env python3

"""Check readline feature."""

from os.path import dirname
from sys import path
from serial import serial_for_url

path.append(dirname(dirname(dirname(dirname(__file__)))))

#pylint: disable-msg=wrong-import-position
from pyftdi import serialext


def main():
    """Verify readline() works with PyFTDI extension."""
    serialext.touch()
    with serial_for_url('ftdi:///1', baudrate=115200) as ser:
        line = ser.readline()
        print(line)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2011-2020, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

from os import environ
from unittest import TestCase, main as ut_main, makeSuite
from pyftdi.jtag import JtagEngine, JtagTool
from pyftdi.bits import BitSequence

#pylint: disable-msg=missing-docstring


# Should match the tested device
JTAG_INSTR = {'SAMPLE': BitSequence('0001', msb=True, length=4),
              'PRELOAD': BitSequence('0001', msb=True, length=4),
              'IDCODE': BitSequence('0100', msb=True, length=4),
              'BYPASS': BitSequence('1111', msb=True, length=4)}


class JtagTestCase(TestCase):

    def setUp(self):
        url = environ.get('FTDI_DEVICE', 'ftdi://ftdi:2232h/1')
        self.jtag = JtagEngine(trst=True, frequency=3E6)
        self.jtag.configure(url)
        self.jtag.reset()
        self.tool = JtagTool(self.jtag)

    def tearDown(self):
        del self.jtag

    def test_idcode_reset(self):
        """Read the IDCODE right after a JTAG reset"""
        self.jtag.reset()
        idcode = self.jtag.read_dr(32)
        self.jtag.go_idle()
        print("IDCODE (reset): 0x%x" % int(idcode))

    def test_idcode_sequence(self):
        """Read the IDCODE using the dedicated instruction"""
        instruction = JTAG_INSTR['IDCODE']
        self.jtag.write_ir(instruction)
        idcode = self.jtag.read_dr(32)
        self.jtag.go_idle()
        print("IDCODE (idcode): 0x%08x" % int(idcode))

    def test_idcode_shift_register(self):
        """Read the IDCODE using the dedicated instruction with
           shift_and_update_register"""
        instruction = JTAG_INSTR['IDCODE']
        self.jtag.change_state('shift_ir')
        retval = self.jtag.shift_and_update_register(instruction)
        print("retval: 0x%x" % int(retval))
        self.jtag.go_idle()
        self.jtag.change_state('shift_dr')
        idcode = self.jtag.shift_and_update_register(BitSequence('0'*32))
        self.jtag.go_idle()
        print("IDCODE (idcode): 0x%08x" % int(idcode))

    def test_bypass_shift_register(self):
        """Test the BYPASS instruction using shift_and_update_register"""
        instruction = JTAG_INSTR['BYPASS']
        self.jtag.change_state('shift_ir')
        retval = self.jtag.shift_and_update_register(instruction)
        print("retval: 0x%x" % int(retval))
        self.jtag.go_idle()
        self.jtag.change_state('shift_dr')
        _in = BitSequence('011011110000'*2, length=24)
        out = self.jtag.shift_and_update_register(_in)
        self.jtag.go_idle()
        print("BYPASS sent: %s, received: %s  (should be left shifted by one)"
              % (_in, out))

    def _test_detect_ir_length(self):
        """Detect the instruction register length"""
        self.jtag.go_idle()
        self.jtag.capture_ir()
        self.tool.detect_register_size()


def suite():
    return makeSuite(JtagTestCase, 'test')


if __name__ == '__main__':
    ut_main(defaultTest='suite')



# Copyright (c) 2008-2016, Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2016, Emmanuel Bouaziz <ebouaziz@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

# this file has not been updated for a while, so coding style needs some love
#pylint: disable-msg=broad-except
#pylint: disable-msg=attribute-defined-outside-init
#pylint: disable-msg=redefined-outer-name
#pylint: disable-msg=invalid-name
#pylint: disable-msg=too-few-public-methods
#pylint: disable-msg=missing-function-docstring
#pylint: disable-msg=missing-class-docstring
#pylint: disable-msg=missing-module-docstring

import errno
import os
import select
import socket
from io import RawIOBase
from serial import (SerialBase, SerialException, portNotOpenError,
                    writeTimeoutError, VERSION as pyserialver)
from ..misc import hexdump


__all__ = ['Serial']


class SerialExceptionWithErrno(SerialException):
    """Serial exception with errno extension"""

    def __init__(self, message, errno=None):
        SerialException.__init__(self, message)
        self.errno = errno


class SocketSerial(SerialBase):
    """Fake serial port redirected to a Unix socket.

       This is basically a copy of the serialposix serial port implementation
       with redefined IO for a Unix socket"""

    BACKEND = 'socket'
    VIRTUAL_DEVICE = True

    PYSERIAL_VERSION = tuple([int(x) for x in pyserialver.split('.')])

    def _reconfigure_port(self):
        pass

    def open(self):
        """Open the initialized serial port"""
        if self._port is None:
            raise SerialException("Port must be configured before use.")
        if self.isOpen():
            raise SerialException("Port is already open.")
        self._dump = False
        self.sock = None
        try:
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            filename = self.portstr[self.portstr.index('://')+3:]
            if filename.startswith('~/'):
                home = os.getenv('HOME')
                if home:
                    filename = os.path.join(home, filename[2:])
            self._filename = filename
            self.sock.connect(self._filename)
        except Exception as e:
            self.close()
            msg = "Could not open port: %s" % (str(e),)
            if isinstance(e, socket.error):
                # pylint: disable-msg=no-member
                raise SerialExceptionWithErrno(msg, e.errno) from e
            raise SerialException(msg) from e
        self._set_open_state(True)
        self._lastdtr = None

    def close(self):
        if self.sock:
            try:
                self.sock.shutdown(socket.SHUT_RDWR)
            except Exception:
                pass
            try:
                self.sock.close()
            except Exception:
                pass
            self.sock = None
        self._set_open_state(False)

    def in_waiting(self):
        """Return the number of characters currently in the input buffer."""
        #pylint: disable-msg=no-self-use
        return 0

    def read(self, size=1):
        """Read size bytes from the serial port. If a timeout is set it may
           return less characters as requested. With no timeout it will block
           until the requested number of bytes is read."""
        if self.sock is None:
            raise portNotOpenError
        read = bytearray()
        if size > 0:
            while len(read) < size:
                ready, _, _ = select.select([self.sock], [], [], self._timeout)
                if not ready:
                    break   # timeout
                buf = self.sock.recv(size-len(read))
                if not buf:
                    # Some character is ready, but none can be read
                    # it is a marker for a disconnected peer
                    raise portNotOpenError
                read += buf
                if self._timeout >= 0 and not buf:
                    break  # early abort on timeout
        return read

    def write(self, data):
        """Output the given string over the serial port."""
        if self.sock is None:
            raise portNotOpenError
        t = len(data)
        d = data
        while t > 0:
            try:
                if self._writeTimeout is not None and self._writeTimeout > 0:
                    _, ready, _ = select.select([], [self.sock], [],
                                                self._writeTimeout)
                    if not ready:
                        raise writeTimeoutError
                n = self.sock.send(d)
                if self._dump:
                    print(hexdump(d[:n]))
                if self._writeTimeout is not None and self._writeTimeout > 0:
                    _, ready, _ = select.select([], [self.sock], [],
                                                self._writeTimeout)
                    if not ready:
                        raise writeTimeoutError
                d = d[n:]
                t = t - n
            except OSError as e:
                if e.errno != errno.EAGAIN:
                    raise

    def flush(self):
        """Flush of file like objects. In this case, wait until all data
           is written."""

    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""

    def reset_output_buffer(self):
        """Clear output buffer, aborting the current output and
        discarding all that is in the buffer."""

    def send_break(self, duration=0.25):
        """Send break condition. Not supported"""

    def _update_break_state(self):
        """Send break condition. Not supported"""

    def _update_rts_state(self):
        """Set terminal status line: Request To Send"""

    def _update_dtr_state(self):
        """Set terminal status line: Data Terminal Ready"""

    def setDTR(self, on=1):
        """Set terminal status line: Data Terminal Ready"""

    @property
    def cts(self):
        """Read terminal status line: Clear To Send"""
        return True

    @property
    def dsr(self):
        """Read terminal status line: Data Set Ready"""
        return True

    @property
    def ri(self):
        """Read terminal status line: Ring Indicator"""
        return False

    @property
    def cd(self):
        """Read terminal status line: Carrier Detect"""
        return False

    # - - platform specific - - - -

    def nonblocking(self):
        """internal - not portable!"""
        if self.sock is None:
            raise portNotOpenError
        self.sock.setblocking(0)

    def dump(self, enable):
        self._dump = enable

    # - - Helpers - -

    def _set_open_state(self, open_):
        if self.PYSERIAL_VERSION < (3, 0):
            self._isOpen = bool(open_)
        else:
            self.is_open = bool(open_)


# assemble Serial class with the platform specifc implementation and the base
# for file-like behavior.
class Serial(SocketSerial, RawIOBase):
    pass



#!/usr/bin/env python3
# Copyright (c) 2019-2021, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause
#
# mock eeprom tests that can be run in CI without a device connected
import logging
from os import environ
from sys import modules, stdout
from unittest import TestCase, TestSuite, SkipTest, makeSuite, main as ut_main
from pyftdi import FtdiLogger
from pyftdi.ftdi import Ftdi
from pyftdi.eeprom import FtdiEeprom
from pyftdi.misc import to_bool, hexdump
from pyftdi.ftdi import FtdiError

VirtLoader = None


class FtdiTestCase(TestCase):
    """Common features for all tests.
    """

    @classmethod
    def setUpClass(cls):
        cls.debug = to_bool(environ.get('FTDI_DEBUG', 'off'), permissive=False)
        cls.url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        cls.loader = None

    @classmethod
    def tearDownClass(cls):
        if cls.loader:
            cls.loader.unload()

    def setUp(self):
        pass


class EepromMirrorTestCase(FtdiTestCase):
    """Test FTDI EEPROM mirror feature (duplicate eeprom data over 2 eeprom
    sectors). Generally this is tested with a virtual eeprom (by setting
    environment variable FTDI_VIRTUAL=on), however you may also test with an
    actual device at your own risk. Note that none of the tests should
    commit any of their eeprom changes
    """

    # manufacturer name string to use in tests
    TEST_MANU_NAME = "MNAME"
    TEST_PROD_NAME = "PNAME"
    TEST_SN = "SN123"

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        if VirtLoader:
            cls.loader = VirtLoader()
            with open(cls.TEST_CONFIG_FILENAME, 'rb') as yfp:
                cls.loader.load(yfp)
        if cls.url == 'ftdi:///1':
            ftdi = Ftdi()
            ftdi.open_from_url(cls.url)
            count = ftdi.device_port_count
            ftdi.close()

    def test_mirror_properties(self):
        """Check FtdiEeprom properties are accurate for a device that can
            mirror
        """
        # properties should work regardless of if the mirror option is set
        # or not
        eeprom = FtdiEeprom()
        eeprom.open(self.url, ignore=True)
        self.assertTrue(eeprom.has_mirroring)
        self.assertEqual(eeprom.size // 2, eeprom.mirror_sector)
        eeprom.close()

        mirrored_eeprom = FtdiEeprom()
        mirrored_eeprom.enable_mirroring(True)
        mirrored_eeprom.open(self.url, ignore=True)
        self.assertTrue(mirrored_eeprom.has_mirroring)
        self.assertEqual(mirrored_eeprom.size // 2,
            mirrored_eeprom.mirror_sector)
        mirrored_eeprom.close()

    def test_mirror_manufacturer(self):
        """Verify manufacturer string is properly duplicated across the 2
            eeprom sectors
        """
        eeprom = FtdiEeprom()
        eeprom.enable_mirroring(True)
        eeprom.open(self.url, ignore=True)
        eeprom.erase()
        eeprom.set_manufacturer_name(self.TEST_MANU_NAME)
        self._check_for_mirrored_eeprom_contents(eeprom)

    def test_mirror_product(self):
        """Verify product string is properly duplicated across the 2 eeprom
            sectors
        """
        eeprom = FtdiEeprom()
        eeprom.enable_mirroring(True)
        eeprom.open(self.url, ignore=True)
        eeprom.erase()
        eeprom.set_product_name(self.TEST_PROD_NAME)
        self._check_for_mirrored_eeprom_contents(eeprom)

    def test_mirror_serial(self):
        """Verify serial string is properly duplicated across the 2 eeprom
            sectors
        """
        eeprom = FtdiEeprom()
        eeprom.enable_mirroring(True)
        eeprom.open(self.url, ignore=True)
        eeprom.erase()
        eeprom.set_serial_number(self.TEST_SN)
        self._check_for_mirrored_eeprom_contents(eeprom)

    def test_varstr_combinations(self):
        """Verify various combinations of var strings are properly duplicated
            across the 2 eeprom sectors
        """
        eeprom = FtdiEeprom()
        eeprom.enable_mirroring(True)
        eeprom.open(self.url, ignore=True)

        # manu + prod str
        eeprom.erase()
        eeprom.set_manufacturer_name(self.TEST_MANU_NAME)
        eeprom.set_product_name(self.TEST_PROD_NAME)
        self._check_for_mirrored_eeprom_contents(eeprom)

        # manu + sn str
        eeprom.erase()
        eeprom.set_manufacturer_name(self.TEST_MANU_NAME)
        eeprom.set_serial_number(self.TEST_SN)
        self._check_for_mirrored_eeprom_contents(eeprom)

        # prod + sn str
        eeprom.erase()
        eeprom.set_manufacturer_name(self.TEST_PROD_NAME)
        eeprom.set_serial_number(self.TEST_SN)
        self._check_for_mirrored_eeprom_contents(eeprom)

        # manu + prod + sn str
        eeprom.erase()
        eeprom.set_manufacturer_name(self.TEST_MANU_NAME)
        eeprom.set_manufacturer_name(self.TEST_PROD_NAME)
        eeprom.set_serial_number(self.TEST_SN)
        self._check_for_mirrored_eeprom_contents(eeprom)

    def test_compute_size_detects_mirror(self):
        """Verify the eeproms internal _compute_size method
            returns the correct bool value when it detects an eeprom mirror
        """
        eeprom = FtdiEeprom()
        eeprom.open(self.url, ignore=True)
        _, mirrored = eeprom._compute_size([])
        self.assertFalse(mirrored)
        test_buf = bytearray(eeprom.size)
        sector_mid = eeprom.size // 2
        for ii in range(sector_mid):
            test_buf[ii] = ii % 255
            test_buf[sector_mid+ii] = test_buf[ii]
        _, mirrored = eeprom._compute_size(bytes(test_buf))
        self.assertTrue(mirrored)

        # change one byte and confirm failure
        test_buf[eeprom.size - 2] = test_buf[eeprom.size - 2] - 1
        _, mirrored = eeprom._compute_size(bytes(test_buf))
        self.assertFalse(mirrored)

    def _check_for_mirrored_eeprom_contents(self, eeprom: FtdiEeprom):
        """Check that contents of the eeprom is identical over the two
            sectors
        """
        sector_size = eeprom.size // 2
        for ii in range(0, sector_size):
            self.assertEqual(eeprom.data[ii], 
                eeprom.data[ii + eeprom.mirror_sector])


class EepromMirrorFt232hTestCase(EepromMirrorTestCase):
    TEST_CONFIG_FILENAME = 'pyftdi/tests/resources/ft232h.yaml'


class EepromMirrorFt2232hTestCase(EepromMirrorTestCase):
    TEST_CONFIG_FILENAME = 'pyftdi/tests/resources/ft2232h.yaml'


class EepromMirrorFt230xTestCase(FtdiTestCase):
    """Test FTDI eeprom with non-mirroring capabilities to ensure it works as
       expected.
    """
    TEST_CONFIG_FILENAME = 'pyftdi/tests/resources/ft230x.yaml'

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        if VirtLoader:
            cls.loader = VirtLoader()
            with open(cls.TEST_CONFIG_FILENAME, 'rb') as yfp:
                cls.loader.load(yfp)
        if cls.url == 'ftdi:///1':
            ftdi = Ftdi()
            ftdi.open_from_url(cls.url)
            count = ftdi.device_port_count
            ftdi.close()
    
    def test_mirror_properties(self):
        """Check FtdiEeprom properties are accurate for a device that can not
           mirror.
        """
        # properties should work regardless of if the mirror option is set
        # or not
        eeprom = FtdiEeprom()
        eeprom.open(self.url, ignore=True)
        self.assertFalse(eeprom.has_mirroring)
        with self.assertRaises(FtdiError):
            eeprom.mirror_sector
        eeprom.close()
        # even if mirroring is enabled, should still stay false
        mirrored_eeprom = FtdiEeprom()
        mirrored_eeprom.enable_mirroring(True)
        mirrored_eeprom.open(self.url, ignore=True)
        self.assertFalse(mirrored_eeprom.has_mirroring)
        with self.assertRaises(FtdiError):
            eeprom.mirror_sector
        mirrored_eeprom.close()

    def test_compute_size_does_not_mirror(self):
        """Verify the eeproms internal _compute_size method returns the correct
           bool value when it detects no mirroring.
        """
        eeprom = FtdiEeprom()
        eeprom.open(self.url, ignore=True)
        _, mirrored = eeprom._compute_size([])
        self.assertFalse(mirrored)
        eeprom.close()

        eeprom = FtdiEeprom()
        eeprom.open(self.url, ignore=False)
        _, mirrored = eeprom._compute_size([])
        self.assertFalse(mirrored)
        eeprom.close()


def suite():
    suite_ = TestSuite()
    suite_.addTest(makeSuite(EepromMirrorFt232hTestCase, 'test'))
    suite_.addTest(makeSuite(EepromMirrorFt2232hTestCase, 'test'))
    suite_.addTest(makeSuite(EepromMirrorFt230xTestCase, 'test'))
    return suite_


def virtualize():
    if not to_bool(environ.get('FTDI_VIRTUAL', 'off')):
        return
    from pyftdi.usbtools import UsbTools
    # Force PyUSB to use PyFtdi test framework for USB backends
    UsbTools.BACKENDS = ('backend.usbvirt', )
    # Ensure the virtual backend can be found and is loaded
    backend = UsbTools.find_backend()
    try:
        # obtain the loader class associated with the virtual backend
        global VirtLoader
        VirtLoader = backend.create_loader()
    except AttributeError:
        raise AssertionError('Cannot load virtual USB backend')


def main():
    import doctest
    doctest.testmod(modules[__name__])
    debug = to_bool(environ.get('FTDI_DEBUG', 'off'))
    if debug:
        formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(levelname)-7s'
                                      ' %(name)-20s [%(lineno)4d] %(message)s',
                                      '%H:%M:%S')
    else:
        formatter = logging.Formatter('%(message)s')
    level = environ.get('FTDI_LOGLEVEL', 'warning').upper()
    try:
        loglevel = getattr(logging, level)
    except AttributeError:
        raise ValueError(f'Invalid log level: {level}')
    FtdiLogger.log.addHandler(logging.StreamHandler(stdout))
    FtdiLogger.set_level(loglevel)
    FtdiLogger.set_formatter(formatter)
    virtualize()
    try:
        ut_main(defaultTest='suite')
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    # Useful environment variables:
    #  FTDI_DEVICE: a specific FTDI URL, default to ftdi:///1
    #  FTDI_LOGLEVEL: a Logger debug level, to define log verbosity
    #  FTDI_DEBUG: to enable/disable debug mode
    #  FTDI_VIRTUAL: to use a virtual device rather than a physical device
    main()



# Copyright (c) 2010-2021 Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import os
import re
import sys

# pip3 install wheel
# pip3 install sphinx-autodoc-typehints sphinx-pypi-upload sphinx_rtd_theme
# python3 setup.py build_sphinx
# sphinx-build -b html ../pyftdi/pyftdi/doc .

topdir = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                      os.pardir, os.pardir))
sys.path.append(topdir)


def read(where, *parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with open(os.path.join(where, *parts), 'rt') as f:
        return f.read()


def find_meta(meta):
    """
    Extract __*meta*__ from meta_file.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        meta_file, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


meta_file = read(topdir, 'pyftdi', '__init__.py')

version = find_meta('version')

needs_sphinx = '2.1'
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx_autodoc_typehints']
templates_path = ['templates']
source_suffix = '.rst'
master_doc = 'index'
project = find_meta('title')
contact = '%s <%s>' % (find_meta('author'), find_meta('email'))
copyright = '2010-2020, %s' % contact
show_authors = True

html_theme = 'sphinx_rtd_theme'
htmlhelp_basename = 'doc'

preamble = r'''
\usepackage{wallpaper}
\usepackage{titlesec}

\titleformat{\chapter}[display]{}{\filleft\scshape\chaptername\enspace\thechapter}{-2pt}{\filright \Huge \bfseries}[\vskip4.5pt\titlerule]
\titleformat{name=\chapter, numberless}[block]{}{}{0pt}{\filright \Huge \bfseries}[\vskip4.5pt\titlerule]

\titlespacing{\chapter}{0pt}{0pt}{1cm}
'''

latex_elements = {
  'papersize': 'a4paper',
  'fncychap': '',  # No Title Page
  'releasename': '',
  'sphinxsetup': 'hmargin={2.0cm,2.0cm}, vmargin={2.5cm,2.5cm}, marginpar=5cm',
  'classoptions': ',openany,oneside',  # Avoid blank page aftre TOC, etc.
  'preamble': preamble,
  'releasename': ''
}

latex_documents = [
  ('index', '%s.tex' % project.lower(),
   '%s Documentation' % project,
   contact, u'manual'),
]

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
latex_toplevel_sectioning = "chapter"

man_pages = [
  ('index', project,
   '%s Documentation' % project,
   [contact], 1)
]

texinfo_documents = [
  ('index', project,
   '%s Documentation' % project,
   contact, '',
   '%s Documentation' % project,
   'Miscellaneous'),
]


def setup(app):
    app.add_css_file('https://fonts.googleapis.com/css?family=Raleway')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2010-2016 Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2010-2016, Neotion
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

#pylint: disable-msg=broad-except

import unittest
from pyftdi.bits import BitSequence, BitZSequence, BitSequenceError


class BitSequenceTestCase(unittest.TestCase):

    def setUp(self):
        self.bs1 = BitSequence(0x01, msb=True, length=8)
        self.bs2 = BitSequence(0x02, msb=True, length=8)
        self.bs3 = BitSequence(0x04, msb=True, length=7)
        self.bs4 = BitSequence(0x04, msb=True, length=11)
        self.bs5 = BitSequence(299999999999998)
        self.bs6 = BitSequence(299999999999999)
        self.bs7 = BitSequence(value='10010101011111')
        self.bzs1 = BitZSequence(0x01, msb=True, length=8)
        self.bzs2 = BitZSequence('0Z1')
        self.bzs3 = BitZSequence('0Z1', length=5)
        self.bzs4 = BitZSequence('0010ZZ010Z1Z11')
        self.bzs5 = BitZSequence(value=[True, False, None, False, False, True])
        self.bzs6 = BitZSequence(value=[True, False, None, False, False, True],
                                 length=len(self.bzs4))

    def test_bitwise_ops(self):
        self.assertEqual(int(BitSequence(0x01, length=8) |
                             BitSequence(0x02, length=8)), 3)
        self.assertEqual(int(BitSequence(0x07, length=8) &
                             BitSequence(0x02, length=8)), 2)
        self.assertEqual(int(BitZSequence(0x01, length=8) |
                             BitSequence(0x02, length=8)), 3)
        self.assertEqual(int(BitSequence(0x07, length=8) &
                             BitZSequence(0x02, length=8)), 2)
        self.assertRaises(BitSequenceError, BitZSequence.__or__,
                          self.bzs4, self.bzs5)
        self.assertRaises(BitSequenceError, BitZSequence.__and__,
                          self.bzs4, self.bzs5)
        self.assertEqual(repr(self.bzs6), '00000000100Z01')
        self.assertEqual(repr(self.bzs6 | self.bzs4), '11Z1Z010ZZ0Z01')
        self.assertEqual(repr(self.bzs6 & self.bzs4), '00Z0Z000ZZ0Z00')
        self.assertEqual(repr(self.bzs4 & self.bs7), '11Z1Z010ZZ0000')
        self.assertEqual(repr(self.bs7 & self.bzs4), '11Z1Z010ZZ0000')
        self.assertEqual(repr(self.bzs4 | self.bs7), '11Z1Z010ZZ1101')
        self.assertEqual(repr(self.bs7 | self.bzs4), '11Z1Z010ZZ1101')
        self.assertEqual(repr(self.bs7.invert()), '00000101010110')
        self.assertEqual(repr(self.bzs4.invert()), '00Z0Z101ZZ1011')
        self.assertLess(self.bs5, self.bs6)
        self.assertLessEqual(self.bs5, self.bs6)
        self.assertLess(self.bs6, self.bs5)
        self.assertLessEqual(self.bs6, self.bs5)

    def test_cmp(self):
        self.assertTrue(self.bs1 == self.bs1)
        self.assertTrue(self.bs1 != self.bs2)
        self.assertTrue(self.bs2 != BitSequence(0x02, msb=True, length=4))
        self.assertTrue(self.bzs2 == self.bzs2)
        self.assertTrue(self.bzs1 != self.bzs2)
        self.assertTrue(self.bs1 == self.bzs1)
        self.assertTrue(self.bzs1 == self.bs1)
        self.assertTrue(self.bzs3 != self.bzs2)
        self.assertNotEqual(self.bzs4, self.bzs5)
        bzs = BitZSequence(self.bs7)
        self.assertTrue(bzs == self.bs7)
        bzs |= BitZSequence('00Z0Z000ZZ0Z00')
        self.assertFalse(bzs == self.bs7)
        self.assertTrue(bzs.matches(self.bs7))

    def test_representation(self):
        self.assertEqual("%s / %r" % (self.bs1, self.bs1),
                         "8: 10000000 / 10000000")
        self.assertEqual("%s / %r" % (self.bs2, self.bs2),
                         "8: 01000000 / 01000000")
        self.assertEqual("%s / %r" % (self.bs3, self.bs3),
                         "7: 0010000 / 0010000")
        self.assertEqual("%s / %r" % (self.bs4, self.bs4),
                         "11: 001 00000000 / 00100000000")
        self.assertEqual("%s / %r" % (self.bs5, self.bs5),
                         "49: 1 00010000 11011001 00110001 01101110 10111111 "
                         "11111110 / 100010000110110010011000101101110101111"
                         "1111111110")
        self.assertEqual("%s / %r" % (self.bs6, self.bs6),
                         "49: 1 00010000 11011001 00110001 01101110 10111111 "
                         "11111111 / 100010000110110010011000101101110101111"
                         "1111111111")

        self.assertEqual(repr(self.bzs4), '11Z1Z010ZZ0100')
        self.assertEqual(repr(self.bzs5), '100Z01')

    def test_init(self):
        self.assertEqual(int(BitSequence([0, 0, 1, 0])), 4)
        self.assertEqual(int(BitSequence((0, 1, 0, 0), msb=True)), 4)
        self.assertEqual(int(BitSequence(4, length=8)), 4)
        self.assertEqual(int(BitSequence(int(4), msb=True, length=8)), 32)
        self.assertEqual(int(BitSequence("0010")), 4)
        self.assertEqual(int(BitSequence("0100", msb=True)), 4)
        bs = BitSequence("0100", msb=True)
        self.assertEqual(bs, BitSequence(bs))
        bssub = BitSequence(bs[1:3])
        self.assertEqual(str(bssub), '2: 10')
        bs[0:3] = '11'
        self.assertEqual(str(bs), '4: 0011')
        bzs = BitZSequence(self.bzs4)
        self.assertEqual(bzs, self.bzs4)
        bs = BitSequence('11111010101001', msb=True)
        bs[8:12] = BitSequence(value='0000')
        self.assertEqual(repr(bs), '11000010101001')
        try:
            bs[8:12] = BitZSequence(value='ZZZZ')
        except BitSequenceError:
            pass
        except Exception as e:
            self.fail("Unexpected exception %s" % e)
        else:
            self.fail("Error was expected")
        bs = BitZSequence('1111101010100111Z1Z010ZZ0100', msb=True)
        bs[8:12] = BitZSequence(value='ZZZZ')
        self.assertEqual(repr(bs), '1111101010100111ZZZZ10ZZ0100')
        bs[8:12] = BitSequence(value='0000')
        self.assertEqual(repr(bs), '1111101010100111000010ZZ0100')
        n = 548521358
        bs = BitSequence(bin(n), msb=True)
        self.assertEqual(int(bs), n)
        bzs = BitZSequence(bin(n), msb=True)
        self.assertEqual(str(bzs), '30: 100000 10110001 11000101 10001110')
        bs = BitSequence(bytes_=[0x44, 0x66, 0xcc], msby=False)
        self.assertEqual(int(bs), 0x4466cc)
        bs = BitSequence(bytes_=(0x44, 0x66, 0xcc), msby=True)
        self.assertEqual(int(bs), 0xcc6644)
        try:
            bs = BitSequence(bytes_=[0x44, 0x666, 0xcc], msby=False)
        except BitSequenceError:
            pass
        except Exception as e:
            self.fail("Unexpected exception %s" % e)
        else:
            self.fail("Error was expected")

    def test_conversion(self):
        bs = BitSequence(0xCA, msb=True, length=8)
        self.assertEqual('%02x' % bs.tobyte(False), '53')
        self.assertEqual('%02x' % bs.tobyte(True), 'ca')
        self.assertEqual(bs, BitSequence(bs.tobyte(True), msb=True, length=8))
        self.assertRaises(BitSequenceError, BitZSequence.__int__, self.bzs5)
        self.assertRaises(BitSequenceError, BitZSequence.tobyte, self.bzs5)
        self.assertRaises(BitSequenceError, BitZSequence.tobytes, self.bzs5)
        bzs = BitZSequence(0xaa)
        self.assertEqual(int(bzs), 0xaa)

    def test_misc(self):
        ba = BitSequence(12, msb=True, length=16)
        bb = BitSequence(12, msb=True, length=14)
        bl = [ba, bb]
        bl.sort(key=int)
        self.assertEqual(str(bl), "[00110000000000, 0011000000000000]")
        self.assertEqual(str(ba.tobytes()), "[48, 0]")
        self.assertEqual(str(ba.tobytes(True)), "[0, 12]")
        self.assertEqual(str(bb.tobytes(True)), "[0, 12]")

        b = BitSequence(length=254)
        b[0:4] = '1111'
        self.assertEqual(
            str(b), '254: 000000 00000000 00000000 00000000 '
            '00000000 00000000 00000000 00000000 00000000 00000000 00000000 '
            '00000000 00000000 00000000 00000000 00000000 00000000 00000000 '
            '00000000 00000000 00000000 00000000 00000000 00000000 00000000 '
            '00000000 00000000 00000000 00000000 00000000 00000000 00001111')
        self.assertEqual(
            str(b.tobytes()),
            '[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, '
            '0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15]')

        b = BitSequence(bytes_=[0xa0, '\x0f', 0x77], msb=False, msby=False)
        self.assertEqual(str(['%02x' % x for x in b.tobytes(False)]),
                         "['a0', '0f', '77']")
        b = BitSequence(bytes_=[0xa0, '\x0f', 0x77], msb=True, msby=True)
        self.assertEqual(str(['%02x' % x for x in b.tobytes(True)]),
                         "['a0', '0f', '77']")
        b = BitSequence(length=7)
        b[6] = '1'
        self.assertEqual(str(b), '7: 1000000')

    def test_rotations(self):
        b = BitSequence('10101110')
        b.lsr(2)
        self.assertEqual(str(b), '8: 01011101')
        b.lsr(10)
        self.assertEqual(str(b), '8: 01010111')
        b.rsr(3)
        self.assertEqual(str(b), '8: 10111010')

    def test_concat(self):
        self.assertEqual(repr(self.bzs4+self.bzs5), '100Z0111Z1Z010ZZ0100')
        self.assertEqual(repr(self.bzs4+self.bs7),
                         '1111101010100111Z1Z010ZZ0100')
        self.assertEqual(repr(self.bs7+self.bzs4),
                         '11Z1Z010ZZ010011111010101001')


def suite():
    return unittest.makeSuite(BitSequenceTestCase, 'test_')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2018, Stephen Goadhouse <sgoadhouse@virginia.edu>
# Copyright (c) 2019, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import logging
import unittest
from doctest import testmod
from os import environ
from sys import modules, stdout
from pyftdi import FtdiLogger
from pyftdi.ftdi import Ftdi
from pyftdi.misc import hexdump, to_bool

#pylint: disable-msg=missing-docstring


class EepromTestCase(unittest.TestCase):
    """FTDI EEPROM access method test case"""

    @classmethod
    def setUpClass(cls):
        """Default values"""
        cls.eeprom_size = int(environ.get('FTDI_EEPROM_SIZE', '256'))
        cls.url = environ.get('FTDI_DEVICE', 'ftdi://ftdi:2232h/1')


    def setUp(self):
        """Open a connection to the FTDI, defining which pins are configured as
           output and input"""
        # out_pins value of 0x00 means all inputs
        out_pins = 0x00
        try:
            ftdi = Ftdi()
            # If you REALLY muck things up, need to use this open_bitbang()
            # function directly and enter vendor and product ID:
            # ftdi.open_bitbang(vendor=0x0403, product=0x6011,
            #                   direction=out_pins)
            ftdi.open_bitbang_from_url(self.url, direction=out_pins)
            self.ftdi = ftdi
        except IOError as exc:
            raise IOError('Unable to open USB port: %s' % str(exc)) from exc

    def tearDown(self):
        """Close the FTDI connection"""
        self.ftdi.close()

    def test_eeprom_read(self):
        """Simple test to demonstrate EEPROM read out.
        """
        ref_data = self.ftdi.read_eeprom(eeprom_size=self.eeprom_size)
        print(hexdump(ref_data))
        # check that the right number of bytes were read
        self.assertEqual(len(ref_data), self.eeprom_size)
        # Pull out actual checksum from EEPROM data
        ck_act = (ref_data[-1] << 8) | ref_data[-2]
        # compute expected checksum value over the EEPROM contents, except
        # the EEPROM word
        ck_expo = self.ftdi.calc_eeprom_checksum(ref_data[:-2])
        self.assertEqual(ck_act, ck_expo)
        maxsize = self.eeprom_size
        # verify access to various data segments
        segments = ((1, 2), (1, 3), (2, 4), (2, 5), (maxsize-8, 8),
                    (maxsize-3, 3), (0, maxsize))
        for start, size in segments:
            chunk = self.ftdi.read_eeprom(start, size, self.eeprom_size)
            self.assertEqual(len(chunk), size)
            self.assertEqual(chunk, ref_data[start:start+size])
        # verify reject access to various invalid data segments
        segments = (-1, 2), (0, maxsize+1), (maxsize-6, maxsize+1)
        for start, size in segments:
            self.assertRaises(ValueError, self.ftdi.read_eeprom, start, size)

    def test_eeprom_write_reject(self):
        """Simple test to demonstrate rejection of invalid EEPROM write
           requests.
        """
        ref_data = self.ftdi.read_eeprom(eeprom_size=self.eeprom_size)
        # check that the right number of bytes were read
        self.assertEqual(len(ref_data), self.eeprom_size)
        # verify reject access to various invalid data segments
        segments = (-1, 2), (0, 257), (250, 7)
        for start, size in segments:
            self.assertRaises(ValueError, self.ftdi.write_eeprom, start,
                              [0] * size, self.eeprom_size)

    def test_eeprom_write(self):
        """Simple test to demonstrate EEPROM write requests.
        """
        self.ftdi.write_eeprom(0x80, b'test', eeprom_size=self.eeprom_size)


def suite():
    suite_ = unittest.TestSuite()
    suite_.addTest(unittest.makeSuite(EepromTestCase, 'test'))
    return suite_


def main():
    if to_bool(environ.get('FTDI_DEBUG', 'off')):
        FtdiLogger.log.addHandler(logging.StreamHandler(stdout))
    level = environ.get('FTDI_LOGLEVEL', 'info').upper()
    try:
        loglevel = getattr(logging, level)
    except AttributeError as exc:
        raise ValueError('Invalid log level: %s' % level) from exc
    FtdiLogger.set_level(loglevel)
    testmod(modules[__name__])
    unittest.main(defaultTest='suite')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2010-2021 Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2010-2016 Neotion
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

#pylint: disable-msg=unused-variable
#pylint: disable-msg=missing-docstring
#pylint: disable-msg=broad-except
#pylint: disable-msg=no-self-use

from codecs import open as codec_open
from setuptools import find_packages, setup
from setuptools.command.build_py import build_py
from distutils.cmd import Command
from distutils.log import DEBUG, INFO
from os import close, getcwd, unlink, walk
from os.path import abspath, dirname, join as joinpath, relpath
from py_compile import compile as pycompile, PyCompileError
from re import split as resplit, search as research
from sys import stderr, exit as sysexit
from tempfile import mkstemp


NAME = 'pyftdi'
PACKAGES = find_packages(where='.')
META_PATH = joinpath('pyftdi', '__init__.py')
KEYWORDS = ['driver', 'ftdi', 'usb', 'serial', 'spi', 'i2c', 'twi', 'rs232',
            'gpio', 'bit-bang']
CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: Other Environment',
    'Natural Language :: English',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: POSIX',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: System :: Hardware :: Hardware Drivers',
]
INSTALL_REQUIRES = [
    'pyusb >= 1.0.0, != 1.2.0',
    'pyserial >= 3.0',
]
TEST_REQUIRES = [
    'ruamel.yaml >= 0.16',
]

HERE = abspath(dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codec_open(joinpath(HERE, *parts), 'rb', 'utf-8') as dfp:
        return dfp.read()


def read_desc(*parts):
    """Read and filter long description
    """
    text = read(*parts)
    text = resplit(r'\.\.\sEOT', text)[0]
    return text


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = research(
        r"(?m)^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        META_FILE
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


class BuildPy(build_py):
    """Override byte-compile sequence to catch any syntax error issue.

       For some reason, distutils' byte-compile when it forks a sub-process
       to byte-compile a .py file into a .pyc does NOT check the success of
       the compilation. Therefore, any syntax error is explictly ignored,
       and no output file is generated. This ends up generating an incomplete
       package w/ a nevertheless successfull setup.py execution.

       Here, each Python file is build before invoking distutils, so that any
       syntax error is catched, raised and setup.py actually fails should this
       event arise.

       This step is critical to check that an unsupported syntax does not end
       up as a 'valid' package from setuptools perspective...
    """

    def byte_compile(self, files):
        for file in files:
            if not file.endswith('.py'):
                continue
            pfd, pyc = mkstemp('.pyc')
            close(pfd)
            try:
                pycompile(file, pyc, doraise=True)
                continue
            except PyCompileError as exc:
                # avoid chaining exceptions
                print(str(exc), file=stderr)
                raise SyntaxError("Cannot byte-compile '%s'" % file)
            finally:
                unlink(pyc)
        super().byte_compile(files)


class CheckStyle(Command):
    """A custom command to check Python coding style."""

    description = 'check coding style'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.announce('checking coding style', level=INFO)
        filecount = 0
        topdir = dirname(__file__) or getcwd()
        for dpath, dnames, fnames in walk(topdir):
            dnames[:] = [d for d in dnames
                         if not d.startswith('.') and d != 'doc']
            for filename in (joinpath(dpath, f)
                             for f in fnames if f.endswith('.py')):
                self.announce('checking %s' % relpath(filename, topdir),
                              level=INFO)
                with open(filename, 'rt') as pfp:
                    for lpos, line in enumerate(pfp, start=1):
                        if len(line) > 80:
                            print('\n  %d: %s' % (lpos, line.rstrip()))
                            raise RuntimeError("Invalid line width '%s'" %
                                               relpath(filename, topdir))
                filecount += 1
        if not filecount:
            raise RuntimeError('No Python file found from "%s"' %
                               topdir)


def main():
    setup(
        cmdclass={
            'build_py': BuildPy,
            'check_style': CheckStyle
        },
        name=NAME,
        description=find_meta('description'),
        license=find_meta('license'),
        url=find_meta('uri'),
        version=find_meta('version'),
        author=find_meta('author'),
        author_email=find_meta('email'),
        maintainer=find_meta('author'),
        maintainer_email=find_meta('email'),
        keywords=KEYWORDS,
        long_description=read_desc('pyftdi/doc/index.rst'),
        packages=PACKAGES,
        scripts=['pyftdi/bin/i2cscan.py',
                 'pyftdi/bin/ftdi_urls.py',
                 'pyftdi/bin/ftconf.py',
                 'pyftdi/bin/pyterm.py'],
        package_dir={'': '.'},
        package_data={'pyftdi': ['*.rst', 'doc/*.rst', 'doc/api/*.rst',
                                 'INSTALL'],
                      'pyftdi.serialext': ['*.rst', 'doc/api/uart.rst']},
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
        test_requires=TEST_REQUIRES,
        python_requires='>=3.7',
    )


if __name__ == '__main__':
    try:
        main()
    except Exception as exc:
        print(exc, file=stderr)
        sysexit(1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2017-2020, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

#pylint: disable-msg=empty-docstring
#pylint: disable-msg=missing-docstring
#pylint: disable-msg=no-self-use

import logging
import unittest
from binascii import hexlify
from doctest import testmod
from os import environ
from sys import modules, stderr, stdout
from time import sleep
from pyftdi import FtdiLogger
from pyftdi.misc import to_bool
from pyftdi.spi import SpiController, SpiIOError


class SpiDataFlashTest:
    """Basic test for a MX25L1606E data flash device selected as CS0,
       SPI mode 0
    """

    def __init__(self):
        self._spi = SpiController(cs_count=3)

    def open(self):
        """Open an SPI connection to a slave"""
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        debug = to_bool(environ.get('FTDI_DEBUG', 'off'))
        self._spi.configure(url, debug=debug)

    def read_jedec_id(self):
        port = self._spi.get_port(0, freq=3E6, mode=0)
        jedec_id = port.exchange([0x9f], 3)
        hex_jedec_id = hexlify(jedec_id).decode()
        print('JEDEC ID:', hex_jedec_id)
        return hex_jedec_id

    def close(self):
        """Close the SPI connection"""
        self._spi.terminate()


class SpiAccelTest:
    """Basic test for an ADXL345 device selected as CS1,
       SPI mode 3
    """

    def __init__(self):
        self._spi = SpiController(cs_count=3)

    def open(self):
        """Open an SPI connection to a slave"""
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        debug = to_bool(environ.get('FTDI_DEBUG', 'off'))
        self._spi.configure(url, debug=debug)

    def read_device_id(self):
        port = self._spi.get_port(1, freq=6E6, mode=3)
        device_id = port.exchange([0x00], 1)
        hex_device_id = hexlify(device_id).decode()
        print('DEVICE ID:', hex_device_id)
        return hex_device_id

    def close(self):
        """Close the SPI connection"""
        self._spi.terminate()


class SpiRfda2125Test:
    """Basic test for a RFDA2125 Digital Controlled Variable Gain Amplifier
       selected as CS2,
       SPI mode 0
    """

    def __init__(self):
        self._spi = SpiController(cs_count=3)
        self._port = None

    def open(self):
        """Open an SPI connection to a slave"""
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        debug = to_bool(environ.get('FTDI_DEBUG', 'off'))
        self._spi.configure(url, debug=debug)
        self._port = self._spi.get_port(2, freq=1E6, mode=0)

    def change_attenuation(self, value):
        if not 0.0 <= value <= 31.5:
            print('Out-of-bound attenuation', file=stderr)
        intval = 63-int(value*2)
        self._port.write(bytes([intval]), 1)

    def close(self):
        """Close the SPI connection"""
        self._spi.terminate()


class SpiTestCase(unittest.TestCase):
    """FTDI SPI driver test case

       Simple test to demonstrate SPI feature.

       Please ensure that the HW you connect to the FTDI port A does match
       the encoded configuration. GPIOs can be driven high or low, so check
       your HW setup before running this test as it might damage your HW.

       Do NOT run this test if you use FTDI port A as an UART or I2C
       bridge -or any unsupported setup!! You've been warned.
    """

    def test_spi1(self):
        spi = SpiDataFlashTest()
        spi.open()
        jedec_id = spi.read_jedec_id()
        self.assertIn(jedec_id, ('c22016', 'bf254a'))
        spi.close()

    def _test_spi2(self):
        spi = SpiAccelTest()
        spi.open()
        device_id = spi.read_device_id()
        self.assertEqual(device_id, 'e5')
        spi.close()

    def _test_spi3(self):
        spi = SpiRfda2125Test()
        spi.open()
        slope = 1
        attenuation = 0.0
        for _ in range(10):
            for _ in range(63):
                attenuation += float(slope)
                print(attenuation/2.0)
                spi.change_attenuation(attenuation/2.0)
                sleep(0.05)  # 50 ms
            slope = -slope
        spi.close()


class SpiGpioTestCase(unittest.TestCase):
    """Basic test for GPIO access w/ SPI mode

       It expects the following I/O setup:

       AD4 connected t0 AC0
       AD5 connected t0 AC1
       AD6 connected t0 AC2
       AD7 connected t0 AC3
    """

    # AD0: SCLK, AD1: MOSI, AD2: MISO, AD3: /CS
    AD_OFFSET = 4
    AC_OFFSET = 8
    PIN_COUNT = 4

    @classmethod
    def setUpClass(cls):
        cls.url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        cls.debug = to_bool(environ.get('FTDI_DEBUG', 'off'))

    def setUp(self):
        self._spi = SpiController(cs_count=1)
        self._spi.configure(self.url, debug=self.debug)
        self._port = self._spi.get_port(0, freq=1E6, mode=0)
        self._io = self._spi.get_gpio()

    def tearDown(self):
        """Close the SPI connection"""
        self._spi.terminate()

    def test_ac_to_ad(self):
        ad_pins = ((1 << self.PIN_COUNT) - 1) << self.AD_OFFSET  # input
        ac_pins = ((1 << self.PIN_COUNT) - 1) << self.AC_OFFSET  # output
        io_pins = ad_pins | ac_pins

        def ac_to_ad(ac_output):
            ac_output &= ac_pins
            ac_output >>= self.AC_OFFSET - self.AD_OFFSET
            return ac_output & ad_pins

        self._io.set_direction(io_pins, ac_pins)
        for ac_pin in range(1 << self.PIN_COUNT):
            ac_out = ac_pin << self.AC_OFFSET
            ad_in = ac_to_ad(ac_out)
            self._io.write(ac_out)
            # random SPI exchange to ensure SPI does not change GPIO
            self._port.exchange([0x00, 0xff], 2)
            buf = self._io.read()
            self.assertEqual(buf, ad_in)
        self.assertRaises(SpiIOError, self._io.write, ad_pins)

    def test_ad_to_ac(self):
        ad_pins = ((1 << self.PIN_COUNT) - 1) << self.AD_OFFSET  # output
        ac_pins = ((1 << self.PIN_COUNT) - 1) << self.AC_OFFSET  # input
        io_pins = ad_pins | ac_pins

        def ad_to_ac(ad_output):
            ad_output &= ad_pins
            ad_output <<= self.AC_OFFSET - self.AD_OFFSET
            return ad_output & ac_pins

        self._io.set_direction(io_pins, ad_pins)
        for ad_pin in range(1 << self.PIN_COUNT):
            ad_out = ad_pin << self.AD_OFFSET
            ac_in = ad_to_ac(ad_out)
            self._io.write(ad_out)
            # random SPI exchange to ensure SPI does not change GPIO
            self._port.exchange([0x00, 0xff], 2)
            buf = self._io.read()
            self.assertEqual(buf, ac_in)
        self.assertRaises(SpiIOError, self._io.write, ac_pins)


class SpiUnalignedTestCase(unittest.TestCase):
    """Basic test for SPI with non 8-bit multiple transfer

       It expects the following I/O setup:

       MOSI (AD1) connected to MISO (AD2)
    """

    @classmethod
    def setUpClass(cls):
        cls.url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        cls.debug = to_bool(environ.get('FTDI_DEBUG', 'off'))

    def setUp(self):
        self._spi = SpiController(cs_count=1)
        self._spi.configure(self.url, debug=self.debug)
        self._port = self._spi.get_port(0, freq=1E6, mode=0)

    def tearDown(self):
        """Close the SPI connection"""
        self._spi.terminate()

    def test_invalid_write(self):
        buf = b'\xff\xff'
        self.assertRaises(ValueError, self._port.write, buf, droptail=8)

    def test_bit_write(self):
        buf = b'\x0f'
        for loop in range(7):
            self._port.write(buf, droptail=loop+1)

    def test_bytebit_write(self):
        buf = b'\xff\xff\x0f'
        for loop in range(7):
            self._port.write(buf, droptail=loop+1)

    def test_invalid_read(self):
        self.assertRaises(ValueError, self._port.read, 1, droptail=8)
        self.assertRaises(ValueError, self._port.read, 2, droptail=8)

    def test_bit_read(self):
        # make MOSI stay to low level, so MISO samples 0
        self._port.write([0x00])
        for loop in range(7):
            data = self._port.read(1, droptail=loop+1)
            self.assertEqual(len(data), 1)
        # make MOSI stay to high level, so MISO samples 1
        self._port.write([0x01])
        for loop in range(7):
            data = self._port.read(1, droptail=loop+1)
            self.assertEqual(len(data), 1)

    def test_bytebit_read(self):
        self._port.write([0x00])
        for loop in range(7):
            data = self._port.read(3, droptail=loop+1)
            self.assertEqual(len(data), 3)
            self.assertEqual(data[-1], 0)
        self._port.write([0x01])
        for loop in range(7):
            data = self._port.read(3, droptail=loop+1)
            self.assertEqual(len(data), 3)

    def test_invalid_duplex(self):
        buf = b'\xff\xff'
        self.assertRaises(ValueError, self._port.exchange, buf,
                          duplex=False, droptail=8)
        self.assertRaises(ValueError, self._port.exchange, buf,
                          duplex=False, droptail=8)
        self.assertRaises(ValueError, self._port.exchange, buf,
                          duplex=True, droptail=8)
        self.assertRaises(ValueError, self._port.exchange, buf,
                          duplex=True, droptail=8)

    def test_bit_duplex(self):
        buf = b'\xcf'
        for loop in range(7):
            data = self._port.exchange(buf, duplex=True, droptail=loop+1)
            self.assertEqual(len(data), 1)
            exp = buf[0] & ~((1<<(loop+1))-1)
            # print(f'{data[0]:08b} {exp:08b}')
            self.assertEqual(data[0], exp)

    def test_bytebit_duplex(self):
        buf = b'\xff\xcf'
        for loop in range(7):
            data = self._port.exchange(buf, duplex=True, droptail=loop+1)
            self.assertEqual(len(data), 2)
            exp = buf[-1] & ~((1<<(loop+1))-1)
            # print(f'{data[-1]:08b} {exp:08b}')
            self.assertEqual(data[0], 0xFF)
            self.assertEqual(data[-1], exp)


class SpiCsForceTestCase(unittest.TestCase):
    """Basic test for exercing direct /CS control.

       It requires a scope or a digital analyzer to validate the signal
       waveforms.
    """

    @classmethod
    def setUpClass(cls):
        cls.url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        cls.debug = to_bool(environ.get('FTDI_DEBUG', 'off'))

    def setUp(self):
        self._spi = SpiController(cs_count=1)
        self._spi.configure(self.url, debug=self.debug)
        self._port = self._spi.get_port(0, freq=1E6, mode=0)

    def tearDown(self):
        """Close the SPI connection"""
        self._spi.terminate()

    def test_cs_default_pulse(self):
        for _ in range(5):
            self._port.force_select()

    def test_cs_long_pulse(self):
        for _ in range(5):
            self._port.force_select(cs_hold=200)

    def test_cs_manual_pulse(self):
        for _ in range(5):
            self._port.force_select(level=False)
            self._port.force_select(level=True)
            # beware that random USB bus access does not allow to create
            # precise delays. This is only the shorter bound, longer one is
            # not defined
            sleep(100e-6)

    def test_cs_pulse_write(self):
        self._port.force_select()
        self._port.write([0x00, 0x01, 0x02])

    def test_cs_default_pulse_rev_clock(self):
        if not self._spi.is_inverted_cpha_supported:
            self.skipTest('FTDI does not support mode 3')
        self._port.set_mode(3)
        for _ in range(5):
            self._port.force_select()


def suite():
    suite_ = unittest.TestSuite()
    # suite_.addTest(unittest.makeSuite(SpiTestCase, 'test'))
    # suite_.addTest(unittest.makeSuite(SpiGpioTestCase, 'test'))
    suite_.addTest(unittest.makeSuite(SpiUnalignedTestCase, 'test'))
    suite_.addTest(unittest.makeSuite(SpiCsForceTestCase, 'test'))
    return suite_


def main():
    testmod(modules[__name__])
    FtdiLogger.log.addHandler(logging.StreamHandler(stdout))
    level = environ.get('FTDI_LOGLEVEL', 'info').upper()
    try:
        loglevel = getattr(logging, level)
    except AttributeError as exc:
        raise ValueError(f'Invalid log level: {level}') from exc
    FtdiLogger.set_level(loglevel)
    unittest.main(defaultTest='suite')


if __name__ == '__main__':
    main()



"""PyUSB virtual USB backend to intercept all USB requests.

   The role of this module is to enable PyFtdi API testing w/o any FTDI
   hardware.
"""

# Copyright (c) 2020-2021, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

#pylint: disable-msg=missing-docstring
#pylint: disable-msg=invalid-name
#pylint: disable-msg=attribute-defined-outside-init
#pylint: disable-msg=too-many-locals
#pylint: disable-msg=too-many-arguments
#pylint: disable-msg=too-many-instance-attributes

from array import array
from binascii import hexlify
from functools import partial
from importlib import import_module
from logging import getLogger
from struct import calcsize as scalc, pack as spack
from sys import version_info
from typing import List, Mapping, Optional, Tuple
from usb.backend import IBackend
from pyftdi.misc import EasyDict
from .consts import USBCONST
from .ftdivirt import VirtFtdi


class VirtEndpoint:
    """Fake USB interface endpoint.
    """

    DESCRIPTOR_FORMAT = '<4BHB'

    def __init__(self, defs: dict,
                 extra: Optional[bytes] = None):
        class EndpointDescriptor(EasyDict):
            pass
        if extra and not isinstance(extra, (bytes, bytearray)):
            raise ValueError('Invalid extra payload')
        self.desc = EndpointDescriptor(
            bLength=scalc(self.DESCRIPTOR_FORMAT),
            bDescriptorType=USBCONST.descriptors.ENDPOINT,
            bEndpointAddress=0,
            bmAttributes=0,
            wMaxPacketSize=64,
            bInterval=0,
            bRefresh=0,
            bSynchAddress=0,
            extra_descriptors=extra or b'')
        self.desc.update(defs)

    def build_strings(self, func):
        func(self.desc)

    def get_length(self) -> int:
        return self.desc.bLength

    def __getattr__(self, name):
        return getattr(self.desc, name)


class VirtInterface:
    """Fake USB configuration interface.
    """

    DESCRIPTOR_FORMAT = '<9B'

    def __init__(self, defs: dict, extra: Optional[bytes] = None):
        class InterfaceDescriptor(EasyDict):
            pass
        if extra and not isinstance(extra, (bytes, bytearray)):
            raise ValueError('Invalid extra payload')
        desc = InterfaceDescriptor(
            bLength=scalc(self.DESCRIPTOR_FORMAT),
            bDescriptorType=USBCONST.descriptors.INTERFACE,
            bInterfaceNumber=0,
            bAlternateSetting=0,
            bNumEndpoints=0,
            bInterfaceClass=0xFF,
            bInterfaceSubClass=0xFF,
            bInterfaceProtocol=0xFF,
            iInterface=0,  # String desc index
            extra_descriptors=extra or  b'')
        desc.update(defs)
        self.alt = 0
        self.altsettings: List[Tuple[VirtInterface,
                                     List[VirtEndpoint]]] = [(desc, [])]

    def add_endpoint(self, endpoint: VirtEndpoint):
        altsetting = self.altsettings[self.alt]
        altsetting[1].append(endpoint)
        altsetting[0].bNumEndpoints = len(altsetting[1])

    def update_number(self, number: int) -> None:
        self.altsettings[self.alt][0].bInterfaceNumber = number

    @property
    def endpoints(self):
        return self.altsettings[self.alt][1]

    def build_strings(self, func):
        for desc, _ in self.altsettings:
            func(desc)
        for _, endpoints in self.altsettings:
            for endpoint in endpoints:
                endpoint.build_strings(func)

    def add_bulk_pair(self):
        endpoints = self.altsettings[self.alt][1]
        desc = {
            'bEndpointAddress': len(endpoints)+1 | USBCONST.endpoints['in'],
            'bmAttributes': USBCONST.endpoint_types['bulk']
        }
        ep = VirtEndpoint(desc)
        self.add_endpoint(ep)
        desc = {
            'bEndpointAddress': len(endpoints)+1 | USBCONST.endpoints['out'],
            'bmAttributes': USBCONST.endpoint_types['bulk']
        }
        ep = VirtEndpoint(desc)
        self.add_endpoint(ep)

    def get_length(self) -> int:
        length = 0
        for desc, endpoints in self.altsettings:
            length += desc.bLength
            for endpoint in endpoints:
                length += endpoint.get_length()
        return length

    @property
    def num_altsetting(self):
        return len(self.altsetting)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.altsettings[item]
        raise IndexError('Invalid alternate setting')

    def __getattr__(self, name):
        return getattr(self.altsettings[self.alt][0], name)


class VirtConfiguration:
    """Fake USB device configuration.
    """

    DESCRIPTOR_FORMAT = '<2BH5B'

    def __init__(self, defs: dict, extra: Optional[bytes] = None):
        class ConfigDescriptor(EasyDict):
            pass
        if extra and not isinstance(extra, (bytes, bytearray)):
            raise ValueError('Invalid extra payload')
        self.desc = ConfigDescriptor(
            bLength=scalc(self.DESCRIPTOR_FORMAT),
            bDescriptorType=USBCONST.descriptors.CONFIG,
            wTotalLength=0,
            bNumInterfaces=0,
            bConfigurationValue=0,
            iConfiguration=0, # string index
            bmAttributes=0x80,  # bus-powered
            bMaxPower=150//2,  # 150 mA
            extra_descriptors=extra or  b'')
        self.desc.update(defs)
        self.interfaces: List[VirtInterface] = []

    def add_interface(self, interface: VirtInterface):
        interface.update_number(len(self.interfaces))
        self.interfaces.append(interface)
        self.desc.bNumInterfaces = len(self.interfaces)

    def build_strings(self, func):
        func(self.desc)
        for iface in self.interfaces:
            iface.build_strings(func)

    def update(self):
        # wTotalLength needs to be updated to the actual length of the
        # sub-objects
        self.desc.wTotalLength = self.get_length()

    def get_length(self) -> int:
        length = self.desc.bLength
        for iface in self.interfaces:
            length += iface.get_length()
        return length

    def __getattr__(self, name):
        return getattr(self.desc, name)


class VirtDevice:
    """Fake USB device.
    """

    DESCRIPTOR_FORMAT = '<2BH4B3H4B'

    DEFAULT_LANGUAGE = 0x0409  # en_US

    def __init__(self, defs: dict, **kwargs):
        class DeviceDescriptor(EasyDict):
            pass
        self.desc = DeviceDescriptor(
            bLength=scalc(self.DESCRIPTOR_FORMAT),
            bDescriptorType=USBCONST.descriptors.DEVICE,
            bcdUSB=0x200,  # USB 2.0
            bDeviceClass=0,
            bDeviceSubClass=0,
            bDeviceProtocol=0,
            bMaxPacketSize0=8,
            idVendor=0,
            idProduct=0,
            bcdDevice=0,
            iManufacturer=0,
            iProduct=0,
            iSerialNumber=0,
            bNumConfigurations=0,  # updated later
            port_number=None,  # unsupported
            port_numbers=None,  # unsupported
            bus=0,
            address=0,
            speed=3,
            eeprom=None)
        self.desc.update(defs)
        self._props = set()
        for key in kwargs:
            # be sure not to allow descriptor override by arbitrary properties
            if key not in defs:
                self.desc[key] = kwargs[key]
                self._props.add(key)
        self.configurations = []
        self.strings = ['']  # slot 0 is reserved
        self._ftdi = VirtFtdi(self.desc.bcdDevice,
                              self.desc.bus, self.desc.address,
                              kwargs.get('eeprom', {}))

    def close(self, freeze: bool = False):
        self._ftdi.close(freeze)

    def terminate(self):
        self._ftdi.terminate()

    def add_configuration(self, config: VirtConfiguration):
        config.update()
        self.configurations.append(config)
        self.desc.bNumConfigurations = len(self.configurations)

    def build(self):
        func = partial(VirtDevice._store_strings, self)
        self.build_strings(func)

    def build_strings(self, func):
        func(self.desc, self._props)
        for config in self.configurations:
            config.build_strings(func)

    @property
    def ftdi(self) -> VirtFtdi:
        return self._ftdi

    @staticmethod
    def _store_strings(obj, desc, ignore=None):
        for dkey in sorted(desc):
            if ignore and dkey in ignore:
                continue
            if isinstance(desc[dkey], str):
                stridx = len(obj.strings)
                obj.strings.append(desc[dkey])
                desc[dkey] = stridx

    def get_string(self, type_: int, index: int) -> str:
        if index == 0:
            if self.desc.get('noaccess', False):
                # simulate unauthorized access to the USB device
                return b''
            # request for list of supported languages
            # only support one
            fmt = '<BBH'
            size = scalc(fmt)
            buf = spack(fmt, size, type_, self.DEFAULT_LANGUAGE)
            return buf
        try:
            value = self.strings[index]
        except IndexError:
            return b''
        ms_str = value.encode('utf-16-le')
        fmt = '<BB'
        size = scalc(fmt) + len(ms_str)
        buf = bytearray(spack('<BB', size, type_))
        buf.extend(ms_str)
        return buf

    def __getattr__(self, name):
        return getattr(self.desc, name)


class VirtDeviceHandle(EasyDict):
    """Device handle wrapper as expected by PyUSB APIs.
    """

    def __init__(self, dev, handle):
        super().__init__(handle=handle,
                         devid=0,
                         device=dev)


class VirtBackend(IBackend):
    """Fake PyUSB backend.

       Implement a subset of PyUSB IBackend interface so that PyFTDI can
       execute w/o a real FTDI HW.
    """

    def __init__(self):
        self.log = getLogger('pyftdi.virt.usb')
        self._devices: List[VirtDevice] = list()
        self._device_handles: Mapping[int, VirtDeviceHandle] = dict()
        self._device_handle_count: int = 0

    def add_device(self, device: VirtDevice):
        self._devices.append(device)

    def flush_devices(self):
        for dev in self._devices:
            dev.terminate()
        self._devices.clear()

    @classmethod
    def create_loader(cls) -> 'VirtLooader':
        """Provide the loader class to configure this virtual backend instance.

           Using this method to retrieve a loader ensure both the virtual
           backend and the loader have been loaded from the same package.

           :return: the VirtLoader class
        """
        # this is a bit circumvoluted, but modules cannot cross-reference
        loader_modname = '.'.join(__name__.split('.')[:-1] + ['loader'])
        loader_mod = import_module(loader_modname)
        VirtLoader = getattr(loader_mod, 'VirtLoader')
        return VirtLoader

    @property
    def devices(self) -> List[VirtDevice]:
        return self._devices

    def get_virtual_ftdi(self, bus: int, address: int) -> VirtFtdi:
        for dev in self._devices:
            if dev.bus == bus and dev.address == address:
                return dev.ftdi
        raise ValueError('No FTDI @ {bus:address}')

    def enumerate_devices(self) -> VirtDevice:
        for dev in self._devices:
            yield dev

    def open_device(self, dev: VirtDevice) -> VirtDeviceHandle:
        self._device_handle_count += 1
        devhdl = VirtDeviceHandle(dev, self._device_handle_count)
        self._device_handles[devhdl.handle] = devhdl
        return devhdl

    def close_device(self, dev_handle: VirtDeviceHandle) -> None:
        del self._device_handles[dev_handle.handle]

    def claim_interface(self, dev_handle: VirtDeviceHandle, intf: int) \
            -> None:
        self.log.info('> claim interface h:%d: if:%d',
                      dev_handle.handle, intf)

    def release_interface(self, dev_handle: VirtDeviceHandle, intf: int) \
            -> None:
        self.log.info('> release interface h:%d: if:%d',
                      dev_handle.handle, intf)

    def get_configuration(self,
                          dev_handle: VirtDeviceHandle) -> VirtConfiguration:
        dev = dev_handle.device
        return dev.configurations[0]

    def set_configuration(self, dev_handle: VirtDeviceHandle,
                          config_value: int) -> None:
        # config_value = ConfigDesc.bConfigurationValue
        pass

    def get_device_descriptor(self, dev: VirtDevice) -> EasyDict:
        return dev.desc

    def get_configuration_descriptor(self, dev: VirtDevice, config: int) \
            -> EasyDict:
        return dev.configurations[config].desc

    def get_interface_descriptor(self, dev: VirtDevice,
                                 intf: int, alt: int, config: int) -> EasyDict:
        cfg = dev.configurations[config]
        iface = cfg.interfaces[intf]
        intf_desc = iface[alt][0]
        return intf_desc

    def get_endpoint_descriptor(self, dev: VirtDevice,
                                ep: int, intf: int, alt: int, config: int) \
            -> EasyDict:
        cfg = dev.configurations[config]
        iface = cfg.interfaces[intf]
        endpoints = iface[alt][1]
        ep_desc = endpoints[ep].desc
        return ep_desc

    def ctrl_transfer(self,
                      dev_handle: VirtDeviceHandle,
                      bmRequestType: int,
                      bRequest: int,
                      wValue: int,
                      wIndex: int,
                      data: array,
                      timeout: int) -> int:
        req_type = USBCONST.dec_req_type(bmRequestType)
        if req_type == 'standard':
            return self._ctrl_standard(dev_handle, bmRequestType, bRequest,
                                       wValue, wIndex, data, timeout)
        if req_type == 'vendor':
            ftdi = dev_handle.device.ftdi
            return ftdi.control(dev_handle, bmRequestType, bRequest,
                                wValue, wIndex, data, timeout)
        self.log.error('Unknown request')
        return 0

    def bulk_write(self, dev_handle: VirtDeviceHandle, ep: int, intf: int,
                   data: array, timeout: int) -> int:
        self.log.debug('> write h:%d ep:%0x02x if:%d, d:%s, to:%d',
                       dev_handle.handle, ep, intf, hexlify(data).decode(),
                       timeout)
        ftdi = dev_handle.device.ftdi
        return ftdi.write(dev_handle, ep, intf, data, timeout)

    def bulk_read(self, dev_handle: VirtDeviceHandle, ep: int, intf: int,
                  buff: array, timeout: int) -> int:
        self.log.debug('> read h:%d ep:0x%02x if:%d, l:%d, to:%d',
                       dev_handle.handle, ep, intf, len(buff), timeout)
        ftdi = dev_handle.device.ftdi
        return ftdi.read(dev_handle, ep, intf, buff, timeout)

    def _ctrl_standard(self,
                       dev_handle: VirtDeviceHandle,
                       bmRequestType: int,
                       bRequest: int,
                       wValue: int,
                       wIndex: int,
                       data: array,
                       timeout: int) -> int:
        req_ctrl = USBCONST.dec_req_ctrl(bmRequestType)
        req_type = USBCONST.dec_req_type(bmRequestType)
        req_rcpt = USBCONST.dec_req_rcpt(bmRequestType)
        req_desc = ':'.join([req_ctrl, req_type, req_rcpt])
        req_name = USBCONST.dec_req_name(bmRequestType, bRequest)
        dstr = (hexlify(data).decode() if USBCONST.is_req_out(bmRequestType)
                else f'({len(data)})')
        self.log.debug('> ctrl_transfer hdl %d, %s, %s, '
                       'val 0x%04x, idx 0x%04x, data %s, to %d',
                       dev_handle.handle, req_desc, req_name,
                       wValue, wIndex, dstr, timeout)
        size = 0
        if req_name == 'get_descriptor':
            desc_idx = wValue & 0xFF
            desc_type = wValue >> 8
            self.log.debug('  %s: 0x%02x',
                           USBCONST.dec_desc_type(desc_type), desc_idx)
            dev = dev_handle.device
            buf = dev.get_string(desc_type, desc_idx)
            size = len(buf)
            data[:size] = array('B', buf)
        else:
            self.log.warning('Unknown request')
        self.log.debug('< (%d) %s', size, hexlify(data[:size]).decode())
        return size


_VirtBackend = VirtBackend()
"""Unique instance of PyUSB virtual backend."""


def get_backend(*_):
    """PyUSB API implementation."""
    return _VirtBackend
